import logging, sys

from .collections import merge_dicts

__all__ = ['setup_logging', 'ColorizeFilter']


def setup_logging(
    name_or_log='',
    reset=False,
    path=None,
    console=False,
    level=logging.DEBUG,
    colorize=False,
    color_by_level=None,
    fmt='%(asctime)s %(levelname)-8s - %(message)s',
    datefmt='%d/%m/%Y %H:%M:%S'
):
    """
    Setup logging (TODO).

    :param name_or_log: TODO
    :param reset: Unregister all previously registered handlers ?
    :type reset: bool
    :param path: TODO
    :type name: str
    :param console: Toggle console output (stdout)
    :type console: bool
    :param level: TODO
    :type level: int
    :param colorize: TODO
    :type colorize: bool
    :param color_by_level: TODO
    :type color_by_level: dict
    :param fmt: TODO
    :type fmt: str
    :param datefmt: TODO
    :type datefmt: str

    **Example usage**

    Setup a console output for logger with name *test*:

    >>> log = setup_logging('a', reset=True, console=True, fmt=None, datefmt=None)
    >>> log.info('this is my info')
    this is my info
    >>> log.debug('this is my debug')
    this is my debug
    >>> log.setLevel(logging.INFO)
    >>> log.debug('this is my hidden debug')
    >>> log.handlers = []  # Remove handlers manually: pas de bras, pas de chocolat !
    >>> log.info('no handlers, no messages ;-)')

    Colorize:

    >>> log = setup_logging('foo', console=True, colorize=True, fmt='%(levelname)-8s - %(message)s')
    >>> log.warning('Attention please!')
    WARNING  - \x1b[33mAttention please!\x1b[0m

    Show how to reset handlers of the logger to avoid duplicated messages (e.g. in doctest):

    >>> log = setup_logging('test', console=True, fmt=None, datefmt=None)
    >>> log = setup_logging('test', console=True, fmt=None, datefmt=None)
    >>> log.info('double message, tu radote pépé')
    double message, tu radote pépé
    double message, tu radote pépé

    Resetting works as expected:

    >>> _ = setup_logging('test', reset=True, console=True, fmt=None, datefmt=None)
    >>> log.info('single message')
    single message

    Logging to a file instead:

    >>> import os, tempfile
    >>>
    >>> with tempfile.NamedTemporaryFile('r') as f:
    ...     log = setup_logging('my-logger', path=f.name)
    ...     log.info('Ceci va probablement jamais être lu!')
    ...     lines = f.read().split(os.linesep)
    ...
    >>> assert 'INFO     - Ceci va probablement jamais être lu!' in lines[0]
    """
    log = logging.getLogger(name_or_log) if isinstance(name_or_log, str) else name_or_log
    if reset:
        log.handlers = []
    log.setLevel(level)
    if colorize:
        log.addFilter(ColorizeFilter(color_by_level=color_by_level))
    if path:
        handler = logging.FileHandler(path)
        handler.setFormatter(logging.Formatter(fmt=fmt, datefmt=datefmt))
        log.addHandler(handler)
    if console:
        handler = logging.StreamHandler(sys.stdout)
        handler.setFormatter(logging.Formatter(fmt=fmt, datefmt=datefmt))
        log.addHandler(handler)
    return log


class ColorizeFilter(logging.Filter):

    color_by_level = {
        logging.DEBUG: 'cyan',
        logging.ERROR: 'red',
        logging.INFO: 'white',
        logging.WARNING: 'yellow'
    }

    def __init__(self, *args, **kwargs):
        self.color_by_level = merge_dicts(
            self.color_by_level,
            kwargs.pop('color_by_level', None) or {})
        super().__init__(*args, **kwargs)

    def filter(self, record):
        record.raw_msg = record.msg
        color = self.color_by_level.get(record.levelno)
        if color:
            import termcolor
            record.msg = termcolor.colored(record.msg, color)
        return True



import os

from pytoolbox import comparison, filesystem, module, validation
from pytoolbox.subprocess import to_args_list
from pytoolbox.types import get_slots
from . import utils

_all = module.All(globals())


class BaseInfo(validation.CleanAttributesMixin, comparison.SlotsEqualityMixin):

    defaults = {}
    attr_name_template = '{name}'

    def __init__(self, info):
        for attr in get_slots(self):
            self._set_attribute(attr, info)

    def _set_attribute(self, name, info):
        """Set attribute `name` value from the `info` or ``self.defaults`` dictionary."""
        value = info.get(self.attr_name_template.format(name=name), self.defaults.get(name))
        setattr(self, name, value)


class Codec(BaseInfo):

    __slots__ = ('long_name', 'name', 'tag', 'tag_string', 'time_base', 'type')

    attr_name_template = 'codec_{name}'

    @staticmethod
    def clean_time_base(value):
        return utils.to_frame_rate(value)


class Format(BaseInfo):

    __slots__ = (
        'bit_rate',
        'duration',
        'filename',
        'format_name',
        'format_long_name',
        'nb_programs',
        'nb_streams',
        'probe_score',
        'size',
        'start_time',
        'tags'
    )

    @staticmethod
    def clean_bit_rate(value):
        return None if value is None else int(value)

    @staticmethod
    def clean_duration(value):
        return None if value is None else float(value)

    @staticmethod
    def clean_nb_programs(value):
        return None if value is None else int(value)

    @staticmethod
    def clean_nb_streams(value):
        return None if value is None else int(value)

    @staticmethod
    def clean_probe_score(value):
        return None if value is None else int(value)

    @staticmethod
    def clean_size(value):
        return None if value is None else int(value)

    @staticmethod
    def clean_start_time(value):
        return None if value is None else float(value)


class Stream(BaseInfo):

    __slots__ = ('avg_frame_rate', 'codec', 'disposition', 'index', 'r_frame_rate', 'time_base')

    codec_class = Codec

    def __init__(self, info):  # pylint:disable=super-init-not-called
        for attr in get_slots(self):
            if attr == 'codec':
                self.codec = self.codec_class(info)
            else:
                self._set_attribute(attr, info)

    @staticmethod
    def clean_avg_frame_rate(value):
        return None if value is None else utils.to_frame_rate(value)

    @staticmethod
    def clean_index(value):
        return None if value is None else int(value)

    @staticmethod
    def clean_r_frame_rate(value):
        return None if value is None else utils.to_frame_rate(value)


class AudioStream(Stream):

    __slots__ = (
        'bit_rate',
        'bits_per_sample',
        'channel_layout',
        'channels',
        'duration',
        'duration_ts',
        'nb_frames',
        'sample_fmt',
        'sample_rate',
        'start_pts',
        'start_time',
        'tags'
    )

    @staticmethod
    def clean_bit_rate(value):
        return None if value is None else int(value)

    @staticmethod
    def clean_bits_per_sample(value):
        return None if value is None else int(value)

    @staticmethod
    def clean_channels(value):
        return None if value is None else int(value)

    @staticmethod
    def clean_duration(value):
        return None if value is None else float(value)

    @staticmethod
    def clean_duration_ts(value):
        return None if value is None else int(value)

    @staticmethod
    def clean_nb_frames(value):
        return None if value is None else int(value)

    @staticmethod
    def clean_sample_rate(value):
        return None if value is None else int(value)

    @staticmethod
    def clean_start_pts(value):
        return None if value is None else int(value)

    @staticmethod
    def clean_start_time(value):
        return None if value is None else float(value)


class SubtitleStream(Stream):

    __slots__ = ('duration', 'duration_ts', 'start_pts', 'start_time', 'tags')

    @staticmethod
    def clean_duration(value):
        return None if value is None else float(value)

    @staticmethod
    def clean_duration_ts(value):
        return None if value is None else int(value)

    @staticmethod
    def clean_start_pts(value):
        return None if value is None else int(value)

    @staticmethod
    def clean_start_time(value):
        return None if value is None else float(value)


class VideoStream(Stream):

    __slots__ = (
        'bit_rate',
        'bit_per_raw_sample',
        'display_aspect_ratio',
        'duration',
        'duration_ts',
        'has_b_frames',
        'height',
        'level',
        'nb_frames',
        'pix_fmt',
        'profile',
        'sample_aspect_ratio',
        'width',
        'tags'
    )

    @staticmethod
    def clean_bit_rate(value):
        return None if value is None else int(value)

    @staticmethod
    def clean_bit_per_raw_sample(value):
        return None if value is None else int(value)

    @staticmethod
    def clean_duration(value):
        return None if value is None else float(value)

    @staticmethod
    def clean_duration_ts(value):
        return None if value is None else int(value)

    @staticmethod
    def clean_height(value):
        return None if value is None else int(value)

    @staticmethod
    def clean_level(value):
        return None if value is None else int(value)

    @staticmethod
    def clean_nb_frames(value):
        return None if value is None else int(value)

    @staticmethod
    def clean_width(value):
        return None if value is None else int(value)

    @property
    def rotation(self):
        tags = self.tags  # pylint:disable=no-member
        return int(0 if tags is None else tags.get('rotate', 0))


class Media(validation.CleanAttributesMixin, comparison.SlotsEqualityMixin):

    __slots__ = ('path', 'options')

    def __init__(self, path, options=None):
        self.path = path
        self.options = options
        self._size = None

    @property
    def directory(self):
        return None if self.is_pipe else os.path.abspath(os.path.dirname(self.path))

    @property
    def is_pipe(self):
        return utils.is_pipe(self.path)

    @property
    def size(self):
        if self._size is None:
            return 0 if self.is_pipe else filesystem.get_size(self.path)
        return self._size

    @size.setter
    def size(self, value):
        self._size = value

    @staticmethod
    def clean_options(value):
        return to_args_list(value)

    def create_directory(self):
        if not self.is_pipe:
            filesystem.makedirs(self.directory)

    def to_args(self, is_input):
        return self.options + (['-i', self.path] if is_input else [self.path])


__all__ = _all.diff(globals())



import functools, os, warnings

from . import console, module, subprocess as py_subprocess

_all = module.All(globals())


class cached_property(object):  # pylint:disable=invalid-name
    """
    Decorator that converts a method with a single self argument into a property
    cached on the instance.

    Optional ``name`` argument allows you to make cached properties of other methods.
    For example ``url=cached_property(get_absolute_url, name='url')``.

    Copyright: Django Project.
    """
    def __init__(self, func, name=None):
        self.func = func
        self.__doc__ = getattr(func, '__doc__')
        self.name = name or func.__name__

    def __get__(self, instance, type=None):  # pylint:disable=redefined-builtin
        if instance is None:
            return self
        res = instance.__dict__[self.name] = self.func(instance)
        return res


def deprecated(func):
    """
    Decorator that can be used to mark functions as deprecated.
    It will result in a warning being emitted when the function is used.

    Credits: https://wiki.python.org/moin/PythonDecoratorLibrary.
    """
    @functools.wraps(func)
    def _deprecated(*args, **kwargs):
        warnings.warn_explicit(
            f'Call to deprecated function {func.__name__}.',
            category=DeprecationWarning,
            filename=func.__code__.co_filename,
            lineno=func.__code__.co_firstlineno + 1)
        return func(*args, **kwargs)
    return _deprecated


class hybridmethod(object):  # pylint:disable=invalid-name
    """
    Decorator that allows a method to be both used as a class method and an instance method.

    Credits: http://stackoverflow.com/questions/18078744/#18078819

    **Example usage**

    >>> class Hybrid(object):
    ...     value = 10
    ...
    ...     def __init__(self):
    ...         self.value = 20
    ...
    ...     @hybridmethod
    ...     def get_value(receiver):
    ...         return receiver.value
    >>> Hybrid.get_value()
    10
    >>> Hybrid().get_value()
    20
    """
    def __init__(self, func):
        self.func = func

    def __get__(self, obj, cls):
        context = obj if obj is not None else cls

        @functools.wraps(self.func)
        def hybrid(*args, **kwargs):
            return self.func(context, *args, **kwargs)

        # optional, mimic methods some more
        hybrid.__func__ = self.func
        hybrid.__self__ = context
        return hybrid


def confirm_it(message, default=False, abort_message='Operation aborted by the user'):
    """Ask for confirmation before calling the decorated function."""
    def _confirm_it(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):  # pylint:disable=inconsistent-return-statements
            if console.confirm(message, default=default):
                return f(*args, **kwargs)
            print(abort_message)
        return wrapper
    return _confirm_it


def disable_iptables(f):
    """
    Stop the iptables service if necessary, execute the decorated function and then reactivate
    iptables if it was previously stopped.
    """
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        try:
            try:
                py_subprocess.cmd('sudo service iptables stop', shell=True)
                print('Disable iptables')
                has_iptables = True
            except Exception:  # pylint:disable=broad-except
                has_iptables = False
            return f(*args, **kwargs)
        finally:
            if has_iptables:  # pylint:disable=used-before-assignment
                print('Enable iptables')
                py_subprocess.cmd('sudo service iptables start', shell=True)
    return wrapper


def root_required(error_message='This script must be run as root.'):
    """Raise an exception if the current user is not root."""
    def _root_required(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            if not os.geteuid() == 0:
                raise RuntimeError(error_message)
            return f(*args, **kwargs)
        return wrapper
    return _root_required


def run_once(f):
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        if wrapper.executed:
            return None
        result = f(*args, **kwargs)
        wrapper.executed = True
        return result
    wrapper.executed = False
    return wrapper


__all__ = _all.diff(globals())



"""
Mix-ins for building your own views.
"""

import os

from django.contrib import messages
from django.core.exceptions import NON_FIELD_ERRORS, ValidationError
from django.db import models
from django.shortcuts import redirect
from django.views.generic import base as generic

from pytoolbox import module
from pytoolbox.django.core import exceptions
from pytoolbox.django.forms import mixins as forms_mixins
from pytoolbox.django.models import utils

_all = module.All(globals())


class AddRequestToFormKwargsMixin(object):
    """Add the view request to the keywords arguments for instantiating the form."""

    def get_form_kwargs(self, *args, **kwargs):
        kwargs = super().get_form_kwargs(*args, **kwargs)
        if self.should_add_request_to_form_kwargs():
            kwargs.update({'request': self.request})
        return kwargs

    def should_add_request_to_form_kwargs(self):
        return issubclass(self.get_form_class(), forms_mixins.RequestMixin)


class BaseModelMultipleMixin(object):

    def get_context_object_name(self, instance_list):
        """Get the name of the item to be used in the context."""
        if self.context_object_name:
            return self.context_object_name
        elif hasattr(instance_list, 'model'):
            return f'{utils.get_base_model(instance_list.model)._meta.model_name}_list'


class BaseModelSingleMixin(object):

    def get_context_object_name(self, instance):
        """Get the name to use for the instance."""
        if self.context_object_name:
            return self.context_object_name
        elif isinstance(instance, models.Model):
            return utils.get_base_model(instance)._meta.model_name


class InitialMixin(object):
    """Add helpers to safely use the URL query string to fill a form with initial values."""

    initials = {}

    def get_initial(self):
        initial = super().get_initial()
        for name, default in self.initials.items():
            self.set_inital(initial, name, default)
        return initial

    def set_inital(self, initial, name, default):
        initial[name] = value = self.request.GET.get(name, default)
        return value

    def set_initial_from_func(self, initial, name, default, func, msg_value, mgs_missing):
        value = self.request.GET.get(name, default)
        if value is not default:
            try:
                value = func(value)
            except ValueError:
                messages.error(self.request, f'{name} - {msg_value}.')
                return None
            except KeyError:
                messages.error(self.request, f'{name} - {mgs_missing}.')
                return None
        initial[name] = value
        return value

    def set_initial_from_model(self, initial, name, default, model, msg_value, mgs_missing):
        value = self.request.GET.get(name, default)
        if value is not default:
            try:
                value = model.objects.for_user(self.request.user).get(pk=value)
            except ValueError:
                messages.error(self.request, f'{name} - {msg_value}.')
                return None
            except model.DoesNotExist:
                messages.error(self.request, f'{name} - {mgs_missing}.')
                return None
        initial[name] = value
        return value


class LoggedCookieMixin(object):
    """Add a "logged" cookie set to "True" if user is authenticated else to "False"."""

    def post(self, *args, **kwargs):
        response = super().post(*args, **kwargs)
        logged = self.request.user.is_authenticated
        response.set_cookie('logged', logged if isinstance(logged, bool) else logged())
        return response


class RedirectMixin(object):
    """Redirect to a page."""

    redirect_view = None

    def dispatch(self, request, *args, **kwargs):
        if self.redirect_view:
            return redirect(self.redirect_view)
        return super().dispatch(request, *args, **kwargs)


class TemplateResponseMixin(generic.TemplateResponseMixin):

    default_template_directory = 'default'

    def get_template_names(self):
        return [self.template_name] if self.template_name else [
            os.path.join(self.template_directory, self.action + '.html'),
            os.path.join(self.default_template_directory, self.action + '.html')
        ]


class ValidationErrorsMixin(object):

    def form_valid(self, form):
        try:
            return super().form_valid(form)
        except ValidationError as e:
            for field, error in exceptions.iter_validation_errors(e):
                if field:
                    form.add_error(field, error)
                else:
                    self._handle_unknown_error(form, error)
            return self.form_invalid(form)

    def _handle_unknown_error(self, form, error):
        form.add_error(NON_FIELD_ERRORS, error)


__all__ = _all.diff(globals())



from selenium import webdriver

from . import common, select, webelements

__all__ = ['Firefox']


class Firefox(common.FindMixin, webdriver.Firefox):

    web_element_classes = {
        'default': webelements.WebElement,
        'select': select.Select
    }

    def _wrap_value(self, value):
        if isinstance(value, tuple(self.web_element_classes.values())):
            return {'ELEMENT': value.id, 'element-6066-11e4-a52e-4f735466cecf': value.id}
        return super()._wrap_value(value)

    def create_web_element(self, element_id):
        element = self.web_element_classes['default'](self, element_id)
        tag_name = getattr(element, 'tag_name', 'default')
        cls = self.web_element_classes.get(tag_name)
        return cls(element) if cls else element



from pytoolbox.selenium import Keys

__all__ = ['BootstrapSliderMixin']


class BootstrapSliderMixin(object):

    component = 'bootstrapSlider'

    @staticmethod
    def clean_value(value):
        return int(value)

    def clear(self):
        pass  # TODO something to do?

    def send_keys(self, *value):
        if len(value) == 1:
            value = self.clean_value(value[0])
            slider_xpath = "..//*[contains(concat(' ', @class, ' '), ' slider-handle ')]"
            slider = next(e for e in self.find_xpath(slider_xpath) if e.is_displayed())
            # TODO detect step and make a loop to reach the target value
            delta = value - self.get_attribute('value')
            if delta > 0:
                slider.send_keys([Keys.RIGHT] * delta)
            elif delta < 0:
                slider.send_keys([Keys.LEFT] * -delta)
            return
        raise NotImplementedError(f'Sending {value} not implemented.')



"""
Some utilities related to the view layer.
"""

from django.http import Http404

__all__ = ['get_model_or_404']


def get_model_or_404(name, *models):
    try:
        return next(m for m in models if m._meta.model_name == name)
    except StopIteration:
        raise Http404()



import itertools

from . import module, throttles
from .types import isiterable

_all = module.All(globals())


def chain(*objects, callback=isiterable):
    """
    Chain the objects, handle non iterable objects gracefully.

    * Set `callback` to a function checking if the object is iterable, defaults to
      :func:`isiterable`.

    **Example usage**

    >>> list(chain(1, 2, '3', [4, 5], {6: 7}))
    [1, 2, '3', 4, 5, 6]
    >>> list(chain(1, 2, '3', callback=lambda a: True))
    Traceback (most recent call last):
        ...
    TypeError: 'int' object is not iterable
    """
    return itertools.chain.from_iterable(o if callback(o) else [o] for o in objects)


def chunk(objects, length, of_type=list):
    """
    Yield successive chunks of defined `length` from `objects`. Last chunk may be smaller.

    **Example usage**

    >>> iterable = iter(list(range(7)))
    >>> list(chunk(iterable, 3))
    [[0, 1, 2], [3, 4, 5], [6]]
    >>> list(chunk(iterable, 3))
    []
    >>> list(chunk((0, 1, (2, 3)), 1, of_type=set))
    [{0}, {1}, {(2, 3)}]
    """
    iterable = iter(objects)
    while True:
        data = of_type(itertools.islice(iterable, 0, length))
        if not data:
            break
        yield data


def extract_single(objects):
    """
    Return the object from objects if there is only one object, else return objects unmodified.

    **Example usage**

    >>> extract_single({6})
    6
    >>> extract_single([10, 2])
    [10, 2]
    >>> extract_single([7])
    7
    >>> extract_single('!')
    '!'
    """
    return next(iter(objects)) if len(objects) == 1 else objects


def throttle(objects, min_delay):
    """
    Consume and skips some objects to yield them at defined `min_delay`. First and last objects are
    always returned. This function is a shortcut for
    ``throttles.TimeThrottle(min_delay).throttle_iterable(objects)``.

    **Example usage**

    >>> import datetime, time
    >>> def slow_range(*args):
    ...     for i in range(*args):
    ...         time.sleep(0.5)
    ...         yield i
    >>> list(throttle(list(range(10)), datetime.timedelta(minutes=1)))
    [0, 9]
    >>> list(throttle(slow_range(3), '00:00:00.2'))
    [0, 1, 2]
    """
    return throttles.TimeThrottle(min_delay).throttle_iterable(objects)


__all__ = _all.diff(globals())



from pytoolbox import collections, exceptions

__all__ = ['Brand']


class Brand(object):

    brands = frozenset([
        'Apple',
        'Asus',
        'Canon',
        'Fujifilm',
        'HP',
        'HTC',
        'Huawei',
        'GoPro',
        'Hewlett-Packard',
        'Kodak',
        'LGE',
        'Nikon',
        'Olympus',
        'Pentax',
        'Samsung',
        'Sigma',
        'Sony',
        'Sony Ericsson',
        'Tamron'
    ])
    clean_map = collections.merge_dicts({b.lower(): b for b in brands}, {
        # maps the group of Exif.Group.Label
        'canoncs': 'Canon',
        'canoncf': 'Canon',
        'HUAWEI': 'Huawei',
        'lg electronics': 'LGE',
        'nikon3': 'Nikon',
        'nikonld2': 'Nikon',
        'nikonld3': 'Nikon',
        'olympus2': 'Olympus',
        'sony1': 'Sony',
        # maps the value of Exif.Image.Make
        'eastman kodak company': 'Kodak',
        'nikon corporation': 'Nikon',
        'olympus imaging corp.': 'Olympus',
        'olympus optical co.,ltd': 'Olympus',
        'samsung techwin': 'Samsung',
        'semc': 'Sony Ericsson',
        'tamron': 'Tamron'
    })

    def __new__(cls, brand):
        return cls.clean(brand)

    @classmethod
    def clean(cls, brand):
        brand = brand.strip() if brand else brand
        if not brand:
            return None
        brand = cls.clean_map.get(brand.lower(), brand)
        if brand not in cls.brands:
            raise exceptions.InvalidBrandError(brand=brand, brands=cls.brands)
        return brand



import errno, itertools, re, select, subprocess, sys, time
from pathlib import Path

from pytoolbox import filesystem, subprocess as py_subprocess
from . import encode, ffprobe  # pylint:disable=unused-import

__all__ = ['FRAME_MD5_REGEX', 'FFmpeg']

FRAME_MD5_REGEX = re.compile(r'[a-z0-9]{32}', re.MULTILINE)


class FFmpeg(object):
    """
    Encode a set of input files input to a set of output files and yields statistics about the
    encoding.
    """

    executable = 'ffmpeg'
    ffprobe_class = ffprobe.FFprobe
    statistics_class = encode.EncodeStatistics

    def __init__(
        self,
        executable=None,
        chunk_read_timeout=0.5,
        encode_poll_delay=0.5,
        encoding='utf-8'
    ):
        self.executable = executable or self.executable
        self.chunk_read_timeout = chunk_read_timeout
        self.encode_poll_delay = encode_poll_delay
        self.encoding = encoding
        self.ffprobe = self.ffprobe_class()

    def __call__(self, *arguments):
        """Call FFmpeg with given arguments (connect stderr to a PIPE)."""
        return py_subprocess.raw_cmd(
            itertools.chain([self.executable], arguments),
            stderr=subprocess.PIPE,
            universal_newlines=True)

    def encode(  # pylint:disable=too-many-locals
        self,
        inputs,
        outputs,
        in_options=None,
        out_options=None,
        create_directories=True,
        process_poll=True,
        process_kwargs=None,
        statistics_kwargs=None
    ):
        """
        Encode a set of input files input to a set of output files and yields statistics about the
        encoding.
        """
        arguments, inputs, outputs, in_options, out_options = \
            self._get_arguments(inputs, outputs, in_options, out_options)

        # Create outputs directories
        if create_directories:
            for output in outputs:
                output.create_directory()

        statistics = self.statistics_class(
            inputs,
            outputs,
            in_options,
            out_options,
            **(statistics_kwargs or {}))

        process = self._get_process(arguments, **(process_kwargs or {}))
        try:
            yield statistics.start(process)
            while True:
                chunk = self._get_chunk(process)
                yield statistics.progress(chunk or '')
                if process_poll:
                    returncode = process.poll()
                    if returncode is not None:
                        break
                if self.encode_poll_delay:
                    time.sleep(self.encode_poll_delay)
            yield statistics.end(returncode)
        except Exception as e:
            traceback = sys.exc_info()[2]
            py_subprocess.kill(process)
            raise e.with_traceback(traceback) if hasattr(e, 'with_traceback') else e

    @staticmethod
    def get_frames_md5_checksum(filename):
        with filesystem.TempStorage() as tmp:
            checksum_filename = tmp.create_tmp_file(return_file=False)
            FFmpeg()('-y', '-i', filename, '-f', 'framemd5', checksum_filename).wait()
            with open(checksum_filename, encoding='utf-8') as f:
                match = FRAME_MD5_REGEX.search(f.read())
            return match.group() if match else None

    def _clean_medias_argument(self, value):
        """
        Return a list of Media instances from passed value.
        Value can be one or multiple instances of string or Media.
        """
        values = [value] if isinstance(value, (str, Path, self.ffprobe.media_class)) else value
        return [self.ffprobe.to_media(v) for v in values] if values else []

    def _get_arguments(self, inputs, outputs, in_options=None, out_options=None):
        """
        Return the arguments for the encoding process.

        * Set inputs to one or multiple strings (paths) or Media instances (with options).
        * Set outputs to one or multiple strings (paths) or Media instances (with options).
        * Set options to a string or a list with the options to put in-between the inputs and
          outputs (legacy API).

        In return you will get a tuple with (arguments, inputs -> list Media, outputs -> list Media,
        options -> list).
        """
        inputs = self._clean_medias_argument(inputs)
        outputs = self._clean_medias_argument(outputs)
        in_options = py_subprocess.to_args_list(in_options)
        out_options = py_subprocess.to_args_list(out_options)
        args = [self.executable, '-y']
        args.extend(in_options)
        for the_input in inputs:
            args.extend(the_input.to_args(is_input=True))
        args.extend(out_options)
        for output in outputs:
            args.extend(output.to_args(is_input=False))
        return args, inputs, outputs, in_options, out_options

    def _get_chunk(self, process):
        select.select([process.stderr], [], [], self.chunk_read_timeout)
        try:
            chunk = process.stderr.read()
            if chunk is None or isinstance(chunk, str):
                return chunk
            return chunk.decode(self.encoding)
        except IOError as e:
            if e.errno != errno.EAGAIN:
                raise
        return None

    @staticmethod
    def _get_process(arguments, **process_kwargs):
        """Return an encoding process with stderr made asynchronous."""
        process = py_subprocess.raw_cmd(
            arguments,
            stderr=subprocess.PIPE,
            close_fds=True,
            **process_kwargs)
        py_subprocess.make_async(process.stderr)
        return process



from .encode import *         # noqa:F401,F403
from .ffmpeg import *         # noqa:F401,F403
from .ffprobe import *        # noqa:F401,F403
from .miscellaneous import *  # noqa:F401,F403
from .utils import *          # noqa:F401,F403



import struct

from fastxor import fast_xor_inplace  # pylint:disable=no-name-in-module

from pytoolbox.network.rtp import RtpPacket

__all__ = ['FecPacket']


class FecPacket(object):  # pylint:disable=too-many-instance-attributes
    """
    This represent a real-time transport protocol (RTP) packet.

    * :rfc:`2733`
    * `Wikipedia (RTP) <http://en.wikipedia.org/wiki/Real-time_Transport_Protocol>`_
    * `Parameters (RTP) <http://www.iana.org/assignments/rtp-parameters/rtp-parameters.xml>`_

    **Packet header**

    .. code-block:: text

         0                   1                   2                   3
         0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        |       SNBase low bits         |        Length recovery        |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        |E| PT recovery |                    Mask                       |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        |                          TS recovery                          |
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
        |N|D|type |index|    Offset     |      NA       |SNBase ext bits|
        +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+

    The constructor will parse input bytes array to fill packet's fields.
    In case of error (e.g. bad version number) the constructor will abort filling fields and
    un-updated fields are set to their corresponding default value.

    :param bytes: Input array of bytes to parse as a RTP packet with FEC payload
    :type bytes: bytearray
    :param length: Amount of bytes to read from the array of bytes
    :type length: int
    :return: Generated RTP packet with SMPTE 2022-1 FEC payload (aka FEC packet)

    **Example usage**

    Testing header fields value (based on packet 3 of capture DCM_FEC_2D_6_10.pcap):

    * 1st row: RTP header, sequence = 37 798
    * 2nd row: FEC header, SN = 50 288, PT recovery = 0, TS recovery = 7850

    >>> header = bytearray.fromhex('806093a6 00000000 00000000 c4700000 80000000 00001eaa 00060a00')
    >>> length = 1344 - len(header)
    >>> print(length)
    1316
    >>> bytes = header + bytearray(length)
    >>> print(len(bytes))
    1344
    >>> fec = FecPacket(bytes, len(bytes))
    >>> assert fec.valid
    >>> print(fec)
    errors                = []
    sequence              = 37798
    algorithm             = XOR
    direction             = COL
    snbase                = 50288
    offset                = 6
    na                    = 10
    L x D                 = 6 x 10
    payload type recovery = 0
    timestamp recovery    = 7850
    length recovery       = 0
    payload recovery size = 1316
    missing               = []

    Testing header fields value (based on packet 5 of capture DCM_FEC_2D_6_10.pcap):

    * 1st row: RTP header, sequence = 63 004
    * 2nd row: FEC header, SN = 50 344, PT recovery = 0, TS recovery = 878

    >>> header = bytearray.fromhex('8060f61c 00000000 00000000 c4a80000 80000000 0000036e 40010600')
    >>> length = 1344 - len(header)
    >>> print(length)
    1316
    >>> bytes = header + bytearray(length)
    >>> print(len(bytes))
    1344
    >>> fec = FecPacket(bytes, len(bytes))
    >>> assert fec.valid
    >>> print(fec)
    errors                = []
    sequence              = 63004
    algorithm             = XOR
    direction             = ROW
    snbase                = 50344
    offset                = 1
    na                    = 6
    L x D                 = 6 x None
    payload type recovery = 0
    timestamp recovery    = 878
    length recovery       = 0
    payload recovery size = 1316
    missing               = []
    """

    # <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< Constants >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

    ER_PAYLOAD_TYPE = 'RTP Header : Payload type must be set to 96'
    ER_EXTENDED = 'SMPTE 2022-1 Header : Extended must be set to one'
    ER_MASK = 'SMPTE 2022-1 Header : Mask must be set to zero'
    ER_N = 'SMPTE 2022-1 Header : N must be set to zero'
    ER_ALGORITHM = 'SMPTE 2022-1 Header : Algorithm must be set to XOR'
    ER_DIRECTION = 'SMPTE 2022-1 Header : Direction must be COL or ROW'
    ER_INDEX = 'SMPTE 2022-1 Header : Index must be set to zero'
    ER_LD = 'SMPTE 2022-1 Header : The following limitation failed : L*D <= 256'
    ER_L = 'SMPTE 2022-1 Header : The following limitation failed : 1 <= L <= 50'
    ER_D = 'SMPTE 2022-1 Header : The following limitation failed : 4 <= D <= 50'
    ER_PAYLOAD = "FEC packet must have a payload"
    ER_ALGORITHM = 'SMPTE 2022-1 Header : Only XOR FEC algorithm is handled'
    ER_VALID_MP2T = 'One of the packets is an invalid RTP packet (+expected MPEG2-TS payload)'
    ER_OFFSET = '(packets) Computed offset is out of range [1..255]'
    ER_SEQUENCE = "One of the packets doesn't verify : sequence = snbase + i * offset, 0<i<na"
    ER_INDEX = 'Unable to get missing media packet index'
    ER_J = 'Unable to find a suitable j e N that satisfy : media_sequence = snbase + j * offset'

    HEADER_LENGTH = 16
    E_MASK = 0x80
    PT_MASK = 0x7f
    N_MASK = 0x80
    D_MASK = 0x40
    T_MASK = 0x38
    T_SHIFT = 3
    I_MASK = 0x07
    SNBL_MASK = 0xffff
    SNBE_SHIFT = 16

    DIRECTION_NAMES = ('COL', 'ROW')
    DIRECTION_RANGE = range(len(DIRECTION_NAMES))  # noqa
    COL, ROW = DIRECTION_RANGE

    ALGORITHM_NAMES = ('XOR', 'Hamming', 'ReedSolomon')
    ALGORITHM_RANGE = range(len(ALGORITHM_NAMES))  # noqa
    XOR, Hamming, ReedSolomon = ALGORITHM_RANGE

    # <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< Properties >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

    @property
    def valid(self):
        """Returns True if this packet is a valid RTP packet (with FEC payload)."""
        return len(self.errors) == 0

    @property
    def errors(self):
        """
        Returns an array containing any errors.

        :return: array of error message(s).

        **Example usage**

        Testing invalid header:

        #TODO >>> fec = FecPacket(bytearray(FecPacket.HEADER_LENGTH-1), FecPacket.HEADER_LENGTH-1)
        #TODO >>> print(fec.errors)
        #TODO ['RTP Header : Version must be set to 2', 'RTP packet must have a payload']
        """
        errors = self._errors[:]
        if not self.extended:
            errors.append(self.ER_EXTENDED)
        if self.mask != 0:
            errors.append(self.ER_MASK)
        if self.n:
            errors.append(self.ER_N)
        if self.algorithm != self.XOR:
            errors.append(self.ER_ALGORITHM)
        if self.direction not in self.DIRECTION_RANGE:
            errors.append(self.ER_DIRECTION)
        if self.index != 0:
            errors.append(self.ER_INDEX)
        if self.payload_size == 0:
            errors.append(self.ER_PAYLOAD)
        if self.L < 1 or self.L > 50:
            errors.append(self.ER_L)
        if self.direction == self.COL and self.L * self.D > 256:
            errors.append(self.ER_LD)
        if self.direction == self.COL and (self.D < 4 or self.D > 50):
            errors.append(self.ER_D)
        return errors

    @property
    def D(self):  # pylint:disable=invalid-name
        """
        Returns the vertical size of the FEC matrix (rows).

        **Example usage**

        >>> packets = [RtpPacket.create(10, 100, RtpPacket.MP2T_PT, bytearray(123)),
        ...            RtpPacket.create(11, 200, RtpPacket.MP2T_PT, bytearray(1234))]
        >>> fec = FecPacket.compute(10, FecPacket.XOR, FecPacket.COL, 1, 2, packets)
        >>> print(fec.D)
        2
        """
        return self.na if self.direction == self.COL else None

    @property
    def L(self):  # pylint:disable=invalid-name
        """
        Returns the horizontal size of the FEC matrix (columns).

        **Example usage**

        >>> from pytoolbox.network.rtp import RtpPacket
        >>> packets = [RtpPacket.create(10, 100, RtpPacket.MP2T_PT, bytearray(123)),
        ...            RtpPacket.create(11, 200, RtpPacket.MP2T_PT, bytearray(1234))]
        >>> fec = FecPacket.compute(6, FecPacket.XOR, FecPacket.ROW, 2, 1, packets)
        >>> print(fec.L)
        2
        """
        return self.offset if self.direction == self.COL else self.na

    @property
    def header_size(self):
        """
        Returns the length (aka size) of the header.

        **Example usage**

        >>> from pytoolbox.network.rtp import RtpPacket
        >>> packets = [RtpPacket.create(10, 100, RtpPacket.MP2T_PT, bytearray(123)),
        ...            RtpPacket.create(11, 200, RtpPacket.MP2T_PT, bytearray(1234))]
        >>> fec = FecPacket.compute(1985, FecPacket.XOR, FecPacket.ROW, 2, 1, packets)
        >>> print(fec.header_size)
        16
        """
        return self.HEADER_LENGTH

    @property
    def payload_size(self):
        """
        Returns the length (aka size) of the payload.

        **Example usage**

        >>> from pytoolbox.network.rtp import RtpPacket
        >>> packets = [RtpPacket.create(10, 100, RtpPacket.MP2T_PT, bytearray(123)),
        ...            RtpPacket.create(11, 200, RtpPacket.MP2T_PT, bytearray(1234))]
        >>> fec = FecPacket.compute(27, FecPacket.XOR, FecPacket.ROW, 2, 1, packets)
        >>> print(fec.payload_size)
        1234
        """
        return len(self.payload_recovery) if self.payload_recovery else 0

    @property
    def header_bytes(self):
        """
        Returns SMPTE 2022-1 FEC header bytes.

        **Example usage**

        >>> from pytoolbox.network.rtp import RtpPacket
        >>> packets = [
        ...     RtpPacket.create(10, 100, RtpPacket.MP2T_PT, bytearray(123)),
        ...     RtpPacket.create(11, 200, RtpPacket.MP2T_PT, bytearray(1234))
        ... ]
        >>> fec = FecPacket.compute(26, FecPacket.XOR, FecPacket.ROW, 2, 1, packets)
        >>> print(fec)
        errors                = []
        sequence              = 26
        algorithm             = XOR
        direction             = ROW
        snbase                = 10
        offset                = 1
        na                    = 2
        L x D                 = 2 x None
        payload type recovery = 0
        timestamp recovery    = 172
        length recovery       = 1193
        payload recovery size = 1234
        missing               = []
        >>> fec_header = fec.header_bytes
        >>> len(fec_header)  # FecPacket.HEADER_LENGTH
        16
        >>> ''.join(' %02x' % b for b in fec_header)
        ' 00 0a 04 a9 80 00 00 00 00 00 00 ac 40 01 02 00'
        >>> fec_header += fec.payload_recovery
        >>> rtp = RtpPacket.create(26, 100, RtpPacket.DYNAMIC_PT, fec_header)
        >>> header = rtp.header_bytes + fec_header
        >>> fec == FecPacket(header, len(header))
        True
        """
        # TODO map type string to enum
        header = bytearray(self.HEADER_LENGTH)
        struct.pack_into(b'!H', header, 0, self.snbase & self.SNBL_MASK)
        struct.pack_into(b'!H', header, 2, self.length_recovery)
        struct.pack_into(b'!I', header, 4, self.mask)
        header[4] = (
            (self.payload_type_recovery & self.PT_MASK)
            + (self.E_MASK if self.extended else 0))
        struct.pack_into(b'!I', header, 8, self.timestamp_recovery)
        header[12] = (
            (self.N_MASK if self.n else 0)
            + (self.D_MASK if self.direction else 0)
            + ((self.algorithm << self.T_SHIFT) & self.T_MASK)
            + (self.index & self.I_MASK))
        header[13] = self.offset
        header[14] = self.na
        header[15] = self.snbase >> self.SNBE_SHIFT
        return header

    @property
    def bytes(self):
        return self.header_bytes + self.payload_recovery

    # <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< Constructor >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

    def __init__(self, data=None, length=0):

        # Fields default values
        self._errors = []
        self.sequence = 0
        self.algorithm = self.XOR
        self.direction = self.COL
        self.snbase = 0
        self.offset = 0
        self.na = 0  # pylint:disable=invalid-name
        self.payload_type_recovery = 0
        self.timestamp_recovery = 0
        self.length_recovery = 0
        self.payload_recovery = []
        # (Unused as defined in SMPTE 2022-1-1)
        self.index = 0
        self.mask = 0
        self.extended = True
        self.n = False  # pylint:disable=invalid-name
        self.missing = []

        if data is not None:
            packet = RtpPacket(data, length)
            self.sequence = packet.sequence
            self._errors = packet.errors[:]
            if len(self._errors) > 0:
                return
            if packet.payload_type != RtpPacket.DYNAMIC_PT:
                self._errors.append(self.ER_PAYLOAD_TYPE)
                return
            self.snbase = (packet.payload[15] * 256 + packet.payload[0]) * 256 + packet.payload[1]
            self.length_recovery = packet.payload[2] * 256 + packet.payload[3]
            self.extended = (packet.payload[4] & self.E_MASK) != 0
            # if not self.extended:
            #     return
            self.payload_type_recovery = packet.payload[4] & self.PT_MASK
            self.mask = (packet.payload[5] * 256 + packet.payload[6]) * 256 + packet.payload[7]
            # if self.mask != 0:
            #     return
            self.timestamp_recovery = (
                ((packet.payload[8] * 256 + packet.payload[9]) * 256 + packet.payload[10]) * 256
                + packet.payload[11])
            self.n = (packet.payload[12] & self.N_MASK) != 0
            # if self.n:
            #     return
            self.direction = (packet.payload[12] & self.D_MASK) >> 6
            self.algorithm = packet.payload[12] & self.T_MASK
            # if self.algorithm != self.XOR:
            #     return
            self.index = packet.payload[12] & self.I_MASK
            # if self.index != 0:
            #     return
            self.offset = packet.payload[13]
            self.na = packet.payload[14]
            self.snbase += packet.payload[15] << 16
            # And finally ... The payload !
            self.payload_recovery = packet.payload[self.HEADER_LENGTH:]

    # <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< Functions >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

    @classmethod
    def compute(cls, sequence, algorithm, direction, L, D, packets):  # pylint:disable=invalid-name
        """
        This method will generate FEC packet's field by applying FEC algorithm to input packets.
        In case of error (e.g. bad version number) the method will abort filling fields and
        un-updated fields are set to their corresponding default value.

        :param sequence: Sequence number of computed FEC packet
        :type sequence: int
        :param algorithm: Name of algorithm used to compute payload recovery from packets payload
        :type algorithm: str
        :param direction: Direction (column or row) of computed FEC packet (see RFC to understand)
        :type direction: str
        :param L: Horizontal size of the FEC matrix (columns)
        :type L: int
        :param D: Vertical size of the FEC matrix (rows)
        :type D: int
        :param packets: Array containing RTP packets to protect
        :type packets: array(RtPacket)

        **Example usage**

        Testing invalid input collection of packets:

        >>> from pytoolbox.network.rtp import RtpPacket
        >>> packets = [
        ...     RtpPacket.create(10, 10, RtpPacket.MP2T_PT, 'a'),
        ...     RtpPacket.create(22, 22, RtpPacket.MP2T_PT, 'b')
        ... ]
        >>> fec = FecPacket.compute(1, FecPacket.XOR, FecPacket.COL, 2, 2, packets)
        Traceback (most recent call last):
            ...
        ValueError: One of the packets doesn't verify : sequence = snbase + i * offset, 0<i<na

        Testing valid input collection of packets:

        >>> packets = [
        ...     RtpPacket.create(10, 10, RtpPacket.MP2T_PT, bytearray('gaga', 'utf-8')),
        ...     RtpPacket.create(14, 14, RtpPacket.MP2T_PT, bytearray('salut', 'utf-8')),
        ...     RtpPacket.create(18, 18, RtpPacket.MP2T_PT, bytearray('12345', 'utf-8')),
        ...     RtpPacket.create(22, 22, RtpPacket.MP2T_PT, bytearray('robot', 'utf-8'))
        ... ]
        >>> fec = FecPacket.compute(2, FecPacket.XOR, FecPacket.COL, 4, 4, packets)
        >>> print(fec)
        errors                = []
        sequence              = 2
        algorithm             = XOR
        direction             = COL
        snbase                = 10
        offset                = 4
        na                    = 4
        L x D                 = 4 x 4
        payload type recovery = 0
        timestamp recovery    = 0
        length recovery       = 1
        payload recovery size = 5
        missing               = []
        >>> print(''.join('%02x:' % x for x in fec.payload_recovery))
        57:5d:5a:4f:35:

        Testing fec packet generation (based on source RTP packets):

        >>> from os import urandom
        >>> from random import randint
        >>> from pytoolbox.network.rtp import RtpPacket
        >>> L = 4
        >>> D = 5
        >>> OFF = 2
        >>> # Generate a [D][L] matrix of randomly generated RTP packets
        >>> matrix = [
        ...     [
        ...         RtpPacket.create(L * j + i, (L * j + i) * 100 + randint(0, 50),
        ...         RtpPacket.MP2T_PT, bytearray(urandom(randint(50, 100))))
        ...         for i in range(L)
        ...     ]
        ...     for j in range(D)
        ... ]
        >>> len(matrix)  # D
        5
        >>> len(matrix[0])  # L
        4

        Retrieve the OFF'th column of the matrix:

        >>> expected_payload_type_recovery = 0
        >>> expected_timestamp_recovery = 0
        >>> expected_lenght_recovery = 0
        >>> expected_payload_recovery = bytearray(100)
        >>> packets = []
        >>> for i in range(D):
        ...     packet = matrix[i][OFF]
        ...     packets.append(packet)
        ...     # Compute expected recovery fields values
        ...     expected_payload_type_recovery ^= packet.payload_type
        ...     expected_timestamp_recovery ^= packet.timestamp
        ...     expected_lenght_recovery ^= packet.payload_size
        ...     for j in range(packet.payload_size):
        ...         expected_payload_recovery[j] ^= packet.payload[j]
        >>> fec = FecPacket.compute(15, FecPacket.XOR, FecPacket.COL, L, D, packets)
        >>> fec.valid
        True
        >>> fec.snbase
        2
        >>> matrix[0][OFF].sequence
        2
        >>> fec.na  # D
        5
        >>> fec.offset  # L
        4
        >>> assert fec.payload_type_recovery == expected_payload_type_recovery
        >>> assert fec.timestamp_recovery == expected_timestamp_recovery
        >>> assert fec.length_recovery == expected_lenght_recovery
        >>> for i in range(fec.payload_size):
        ...     if fec.payload_recovery[i] != expected_payload_recovery[i]:
        ...         print('Payload recovery test failed with i = ' + i)
        """
        # Fields default values
        fec = cls()
        fec.sequence = sequence
        if algorithm not in cls.ALGORITHM_RANGE:
            raise ValueError('algorithm is not a valid FEC algorithm')
        if direction not in cls.DIRECTION_RANGE:
            raise ValueError('direction is not a valid FEC direction')
        fec.algorithm = algorithm
        fec.direction = direction
        if fec.direction == cls.COL:
            fec.na = D
            fec.offset = L
        else:
            fec.na = L
            fec.offset = 1
        if fec.algorithm != cls.XOR:
            raise NotImplementedError(cls.ER_ALGORITHM)
        if len(packets) != fec.na:
            raise ValueError(f'packets must contain exactly {fec.na} packets')
        fec.snbase = packets[0].sequence

        # Detect maximum length of packets payload and check packets validity
        size = i = 0
        for packet in packets:
            if not packet.validMP2T:
                raise ValueError(cls.ER_VALID_MP2T)
            if packet.sequence != (fec.snbase + i * fec.offset) & RtpPacket.S_MASK:
                raise ValueError(cls.ER_SEQUENCE)

            size = max(size, packet.payload_size)
            i += 1

        # Create payload recovery field according to size/length
        fec.payload_recovery = bytearray(size)

        # Compute FEC packet's fields based on input packets
        for packet in packets:

            # Update (...) recovery fields by xor'ing corresponding fields of all packets
            fec.payload_type_recovery ^= packet.payload_type
            fec.timestamp_recovery ^= packet.timestamp
            fec.length_recovery ^= packet.payload_size

            # Update payload recovery by xor'ing all packets payload
            payload = packet.payload
            if len(packet.payload) < size:
                payload = payload + bytearray(size - len(packet.payload))

            fast_xor_inplace(fec.payload_recovery, payload)
            # NUMPY fec.payload_recovery = bytearray(
            #     numpy.bitwise_xor(fec.payload_recovery, payload))
            # XOR LOOP for i in xrange(min(size, len(packet.payload))):
            # XOR LOOP     fec.payload_recovery[i] ^= packet.payload[i]
        return fec

    def compute_j(self, media_sequence):
        """
        TODO
        """
        delta = media_sequence - self.snbase
        if delta < 0:
            delta += RtpPacket.S_MASK + 1
        if delta % self.offset != 0:
            return None
        return int(delta / self.offset)

    def set_missing(self, media_sequence):
        """
        Register a protected media packet as missing.

        **Example usage**

        >>> packets = [
        ...     RtpPacket.create(65530, 65530, RtpPacket.MP2T_PT, bytearray('gaga', 'utf-8')),
        ...     RtpPacket.create(65533, 65533, RtpPacket.MP2T_PT, bytearray('salut', 'utf-8')),
        ...     RtpPacket.create(    0,     0, RtpPacket.MP2T_PT, bytearray('12345', 'utf-8')),
        ...     RtpPacket.create(    3,     3, RtpPacket.MP2T_PT, bytearray('robot', 'utf-8'))
        ... ]
        >>> fec = FecPacket.compute(4, FecPacket.XOR, FecPacket.COL, 3, 4, packets)
        >>> print(fec)
        errors                = []
        sequence              = 4
        algorithm             = XOR
        direction             = COL
        snbase                = 65530
        offset                = 3
        na                    = 4
        L x D                 = 3 x 4
        payload type recovery = 0
        timestamp recovery    = 4
        length recovery       = 1
        payload recovery size = 5
        missing               = []

        Testing that bad input values effectively throws an exception:

        >>> fec.set_missing(fec.snbase + fec.offset + 1)
        Traceback (most recent call last):
            ...
        ValueError: Unable to find a suitable j e N that satisfy : media_sequence = snbase + j * ...
        >>> fec.set_recovered(-1)
        Traceback (most recent call last):
            ...
        ValueError: Unable to find a suitable j e N that satisfy : media_sequence = snbase + j * ...

        Testing set / get of a unique missing value:

        >>> fec.set_missing(0)
        2
        >>> fec.missing[0]
        0
        >>> len(fec.missing)
        1

        Testing simple recovery of a unique value:

        >>> fec.set_recovered(0)
        2
        >>> len(fec.missing)
        0

        Testing set / get of multiple missing values (including re-setting of a value):

        >>> fec.set_missing(3)
        3
        >>> fec.set_missing(3)
        3
        >>> len(fec.missing)
        1
        >>> fec.set_missing(fec.snbase + fec.offset)
        1
        >>> fec.set_missing(fec.snbase)
        0
        >>> len(fec.missing)
        3
        >>> fec.missing
        [3, 65533, 65530]

        Testing re-recovery of a value:

        >>> fec.set_recovered(3)
        3
        >>> fec.set_recovered(3)
        Traceback (most recent call last):
            ...
        ValueError: list.remove(x): x not in list
        >>> fec.missing
        [65533, 65530]
        """
        j = self.compute_j(media_sequence)
        if j is None:
            raise ValueError(self.ER_J)
        if media_sequence not in self.missing:
            self.missing.append(media_sequence)
        return j

    def set_recovered(self, media_sequence):
        """
        TODO
        """
        j = self.compute_j(media_sequence)
        if j is None:
            raise ValueError(self.ER_J)
        self.missing.remove(media_sequence)
        return j

    def __eq__(self, other):
        """
        Test equality field by field !
        """
        return (
            isinstance(other, self.__class__)
            and self.sequence == other.sequence and self.algorithm == other.algorithm
            and self.direction == other.direction and self.snbase == other.snbase
            and self.offset == other.offset and self.na == other.na
            and self.payload_type_recovery == other.payload_type_recovery
            and self.length_recovery == other.length_recovery
            and self.payload_recovery == other.payload_recovery)

    def __str__(self):
        """
        Returns a string containing a formated representation of the packet fields.

        **Example usage**

        >>> packets = [RtpPacket.create(10, 100, RtpPacket.MP2T_PT, bytearray(10)),
        ...            RtpPacket.create(11, 200, RtpPacket.MP2T_PT, bytearray(5)),
        ...            RtpPacket.create(12, 300, RtpPacket.MP2T_PT, bytearray(7)),
        ...            RtpPacket.create(13, 400, RtpPacket.MP2T_PT, bytearray(10))]
        >>> fec = FecPacket.compute(12, FecPacket.XOR, FecPacket.ROW, 4, 1, packets)
        >>> print(fec)
        errors                = []
        sequence              = 12
        algorithm             = XOR
        direction             = ROW
        snbase                = 10
        offset                = 1
        na                    = 4
        L x D                 = 4 x None
        payload type recovery = 0
        timestamp recovery    = 16
        length recovery       = 2
        payload recovery size = 10
        missing               = []

        >>> fec = FecPacket.compute(14, FecPacket.XOR, FecPacket.COL, 1, 4, packets)
        >>> print(fec)
        errors                = []
        sequence              = 14
        algorithm             = XOR
        direction             = COL
        snbase                = 10
        offset                = 1
        na                    = 4
        L x D                 = 1 x 4
        payload type recovery = 0
        timestamp recovery    = 16
        length recovery       = 2
        payload recovery size = 10
        missing               = []
        """
        return f"""errors                = {self.errors}
sequence              = {self.sequence}
algorithm             = {self.ALGORITHM_NAMES[self.algorithm]}
direction             = {self.DIRECTION_NAMES[self.direction]}
snbase                = {self.snbase}
offset                = {self.offset}
na                    = {self.na}
L x D                 = {self.L} x {self.D}
payload type recovery = {self.payload_type_recovery}
timestamp recovery    = {self.timestamp_recovery}
length recovery       = {self.length_recovery}
payload recovery size = {self.payload_size}
missing               = {self.missing}"""



import errno, inspect, json, os, pickle, shutil

from . import filesystem, module
from .private import ObjectId
from .types import get_slots

_all = module.All(globals())


# Data -> File -------------------------------------------------------------------------------------

def to_file(
    path,
    data=None,
    pickle_data=None,
    binary=False,
    safe=False,
    backup=False,
    makedirs=False
):
    """
    Write some data to a file, can be safe (tmp file -> rename), may create a backup before any
    write operation. Return the name of the backup path or None.

    **Example usage**

    In-place write operation:

    >>> to_file('/tmp/to_file', data='bonjour') is None
    True
    >>> open('/tmp/to_file').read()
    'bonjour'

    No backup is created if the destination file does not exist:

    >>> from pytoolbox import filesystem
    >>> _ = filesystem.remove('/tmp/to_file')
    >>> to_file('/tmp/to_file', data='bonjour', backup=True) is None
    True

    In-place write operation after having copied the file into a backup:

    >>> to_file('/tmp/to_file', data='ça va ?', backup=True)
    '/tmp/to_file.bkp'
    >>> open('/tmp/to_file.bkp').read()
    'bonjour'
    >>> open('/tmp/to_file').read()
    'ça va ?'

    The most secure, do a backup, write into a temporary file, and rename the temporary file to the
    destination:

    >>> to_file('/tmp/to_file', data='oui et toi ?', safe=True, backup=True)
    '/tmp/to_file.bkp'
    >>> open('/tmp/to_file.bkp').read()
    'ça va ?'
    >>> open('/tmp/to_file').read()
    'oui et toi ?'

    The content of the destination is not broken if the write operation fails:

    >>> to_file('/tmp/to_file', data=to_file, safe=True)
    Traceback (most recent call last):
        ...
    TypeError: write() argument must be str, not function
    >>> open('/tmp/to_file').read()
    'oui et toi ?'
    """
    if makedirs:
        filesystem.makedirs(os.path.dirname(path))
    if backup:
        backup_path = f'{path}.bkp'
        try:
            shutil.copy2(path, backup_path)
        except IOError as e:
            if e.errno != errno.ENOENT:
                raise
            backup_path = None
    write_path = f'{path}.tmp' if safe else path
    with open(write_path, 'wb' if binary else 'w') as f:
        if data is not None:
            f.write(data)
        if pickle_data is not None:
            pickle.dump(pickle_data, f)
    if safe:
        os.rename(write_path, path)
    return backup_path if backup else None


# Object <-> Pickle file ---------------------------------------------------------------------------

class PickleableObject(object):
    """An :class:`object` serializable/deserializable by :mod:`pickle`."""
    @classmethod
    def read(cls, path, store_path=False, create_if_error=False, **kwargs):
        """Return a deserialized instance of a pickleable object loaded from a file."""
        try:
            with open(path, 'rb') as f:
                the_object = pickle.load(f)
        except Exception:  # pylint:disable=broad-except
            if not create_if_error:
                raise
            the_object = cls(**kwargs)
            the_object.write(path, store_path=store_path)
        if store_path:
            the_object._pickle_path = path  # pylint:disable=all
        return the_object

    def write(self, path=None, store_path=False, safe=False, backup=False, makedirs=False):
        """Serialize `self` to a file, excluding the attribute `_pickle_path`."""
        pickle_path = getattr(self, '_pickle_path', None)
        path = path or pickle_path
        if path is None:
            raise ValueError('A path must be specified')
        try:
            if pickle_path:
                del self._pickle_path
            to_file(
                path,
                pickle_data=self,
                binary=True,
                safe=safe,
                backup=backup,
                makedirs=makedirs)
        finally:
            if store_path:
                self._pickle_path = path  # pylint:disable=attribute-defined-outside-init
            elif pickle_path:
                self._pickle_path = pickle_path  # pylint:disable=attribute-defined-outside-init


# Object <-> JSON string ---------------------------------------------------------------------------

# http://stackoverflow.com/questions/6255387/mongodb-object-serialized-as-json
class SmartJSONEncoderV1(json.JSONEncoder):

    def default(self, obj):  # pylint:disable=arguments-differ
        if ObjectId is not None and isinstance(obj, ObjectId):
            return str(obj)
        if hasattr(obj, '__dict__'):
            return obj.__dict__
        return super().default(obj)


class SmartJSONEncoderV2(json.JSONEncoder):

    def default(self, obj):  # pylint:disable=arguments-differ
        if isinstance(obj, ObjectId):
            return str(obj)
        attributes = {}
        for attr in inspect.getmembers(obj):
            if inspect.isroutine(attr[1]) or inspect.isbuiltin(attr[1]) or attr[0].startswith('__'):
                continue
            if hasattr(obj.__dict__, attr[0]) and not isinstance(attr[1], property):
                continue
            attributes[attr[0]] = attr[1]
        return attributes


def object_to_json(obj, include_properties, **kwargs):
    """
    Serialize an :class:`object` to a JSON string.
    Use one of the *smart* JSON encoder of this module.

    * Set include_properties to True to also include the properties of `obj`.
    * Set kwargs with any argument of the function :mod:`json`.dumps excepting cls.

    **Example usage**

    >>> class Point(object):
    ...     def __init__(self, x=0, y=0):
    ...         self.x = x
    ...         self.y = y
    ...     @property
    ...     def z(self):
    ...         return self.x + self.y
    >>> p1 = Point(x=16, y=-5)
    >>> object_to_json(p1, include_properties=False, sort_keys=True)
    '{"x": 16, "y": -5}'
    >>> object_to_json(p1, include_properties=True, sort_keys=True)
    '{"x": 16, "y": -5, "z": 11}'
    >>> print(object_to_json(p1, include_properties=True, sort_keys=True, indent=4))
    {
        "x": 16,
        "y": -5,
        "z": 11
    }
    """
    return json.dumps(
        obj,
        cls=(SmartJSONEncoderV2 if include_properties else SmartJSONEncoderV1),
        **kwargs)


def json_to_object(cls, json_string, inspect_constructor):
    """
    Deserialize the JSON string `json_string` to an instance of `cls`.

    Set `inspect_constructor` to True to filter input dictionary to avoid sending unexpected keyword
    arguments to the constructor (`__init__`) of `cls`.
    """
    return dict_to_object(cls, json.loads(json_string), inspect_constructor)


def jsonfile_to_object(cls, path_or_file, inspect_constructor):
    """
    Load and deserialize the JSON string stored in a file `path` to an instance of `cls`.

    .. warning::

        Class constructor is responsible of converting attributes to instances of classes with
        `dict_to_object`.

    **Example usage**

    Define the sample class, instantiate it and serialize it to a file:

    >>> import os
    >>> from pytoolbox.unittest import asserts
    >>>
    >>> class Point(object):
    ...     def __init__(self, name=None, x=0, y=0):
    ...         self.name = name
    ...         self.x = x
    ...         self.y = y
    >>> p1 = Point(name='My point', x=10, y=-5)
    >>> with open('test.json', 'w', encoding='utf-8') as f:
    ...     f.write(object_to_json(p1, include_properties=False))
    38

    Deserialize the freshly saved file:

    >>> p2 = jsonfile_to_object(Point, 'test.json', inspect_constructor=False)
    >>> asserts.dict_equal(p1.__dict__, p2.__dict__)
    >>> p2 = jsonfile_to_object(Point, open('test.json'), inspect_constructor=False)
    >>> asserts.dict_equal(p1.__dict__, p2.__dict__)
    >>> os.remove('test.json')
    """
    if isinstance(path_or_file, str):
        f = open(path_or_file, 'r')
    else:
        f = path_or_file
    return json_to_object(cls, f.read(), inspect_constructor)


class JsoneableObject(object):
    """
    An :class:`object` serializable/deserializable by :mod:`json`.

    .. warning::

        Class constructor is responsible of converting attributes to instances of classes with
        `dict_to_object`.

    Convert-back from JSON strings containing extra parameters:

    >>> from pytoolbox.unittest import asserts
    >>>
    >>> class User(object):
    ...     def __init__(self, first_name, last_name):
    ...         self.first_name, self.last_name = first_name, last_name
    ...     @property
    ...     def name(self):
    ...         return '{0} {1}'.format(self.first_name, self.last_name)
    >>>
    >>> class Media(JsoneableObject):
    ...     def __init__(self, author, title):
    ...         self.author = dict_to_object(
    ...             User, author, True) if isinstance(author, dict) else author
    ...         self.title = title

    Sounds good:

    >>> media = Media(User('Andrés', 'Revuelta'), 'Vacances à Phucket')
    >>> media_json = media.to_json(include_properties=False)
    >>> media_back = Media.from_json(media_json, inspect_constructor=True)
    >>> isinstance(media_back.author, User)
    True
    >>> asserts.dict_equal(media_back.author.__dict__, media.author.__dict__)

    A second example handling extra arguments by using ``**kwargs`` (a.k.a the dirty way):

    >>> class User(object):
    ...     def __init__(self, first_name, last_name, **kwargs):
    ...         self.first_name, self.last_name = first_name, last_name
    ...     @property
    ...     def name(self):
    ...         return '{0} {1}'.format(self.first_name, self.last_name)
    >>>
    >>> class Media(JsoneableObject):
    ...     def __init__(self, author, title, **kwargs):
    ...         self.author = User(**author) if isinstance(author, dict) else author
    ...         self.title = title

    Sounds good:

    >>> media = Media(User('Andrés', 'Revuelta'), 'Vacances à Phucket')
    >>> media_json = media.to_json(include_properties=True)
    >>> media_back = Media.from_json(media_json, inspect_constructor=False)
    >>> isinstance(media_back.author, User)
    True
    >>> asserts.dict_equal(media_back.author.__dict__, media.author.__dict__)
    """

    @classmethod
    def read(cls, path, store_path=False, inspect_constructor=True):
        """Return a deserialized instance of a jsoneable object loaded from a file."""
        with open(path, encoding='utf-8') as f:
            the_object = dict_to_object(cls, json.loads(f.read()), inspect_constructor)
            if store_path:
                the_object._json_path = path  # pylint:disable=all
            return the_object

    def write(self, path=None, include_properties=False, safe=False, backup=False, makedirs=False,
              **kwargs):
        """Serialize `self` to a file, excluding the attribute `_json_path`."""
        if path is None and hasattr(self, '_json_path'):
            path = self._json_path
            try:
                del self._json_path
                to_file(
                    path,
                    data=object_to_json(self, include_properties, **kwargs),
                    binary=False,
                    safe=safe,
                    backup=backup,
                    makedirs=makedirs)
            finally:
                self._json_path = path  # pylint:disable=attribute-defined-outside-init
        elif path is not None:
            to_file(
                path,
                data=object_to_json(self, include_properties, **kwargs),
                binary=False,
                safe=safe,
                backup=backup,
                makedirs=makedirs)
        else:
            raise ValueError('A path must be specified')

    def to_json(self, include_properties, **kwargs):
        """Serialize this instance to a JSON string."""
        return object_to_json(self, include_properties, **kwargs)

    @classmethod
    def from_json(cls, json_string, inspect_constructor):
        """Deserialize a JSON string to an instance of `JsoneableObject`."""
        return dict_to_object(cls, json.loads(json_string), inspect_constructor)


# Object <-> Dictionary ----------------------------------------------------------------------------

def object_to_dict(
    obj, schema, depth=0,
    callback=lambda o, s, d: (o, s),
    iterable_callback=lambda o, s, d: list
):
    """
    Convert an :class:`object` to nested python lists and dictionaries to follow given schema.

    * Schema callback makes it possible to dynamically tweak obj and schema!
    * Iterable callback makes it possible to dynamically tweak iterable objects container type!

    **Example usage**

    Define the sample class and convert some instances to a nested structure:

    >>> from pytoolbox.unittest import asserts
    >>>
    >>> class Point(object):
    ...     def __init__(self, name, x, y, p):
    ...         self.name = name
    ...         self.x = x
    ...         self.y = y
    ...         self.p = p
    ...     @property
    ...     def z(self):
    ...         return self.x - self.y

    Simple example:

    >>> SCHEMA = {
    ...     'name': 'name',
    ...     'x': 'x',
    ...     'zed': 'z',
    ...     'p': [{
    ...         'name': 'name',
    ...         'stuff': lambda p: 42,
    ...         'p': {
    ...             'name': 'name'
    ...         }
    ...     }]
    ... }
    ...
    >>> asserts.dict_equal(
    ...     object_to_dict(
    ...         Point('p1', 5, 2, [
    ...             Point('p2', 3, 4, None),
    ...             Point('p3', 7, 0, Point('p4', 0, 0, None))
    ...         ]),
    ...         SCHEMA),
    ...     {
    ...         'name': 'p1', 'x': 5, 'zed': 3,
    ...         'p': [
    ...             {'name': 'p2', 'stuff': 42, 'p': None},
    ...             {'name': 'p3', 'stuff': 42, 'p': {'name': 'p4'}}
    ...         ]
    ...     })

    This serializer can optionally adapt container type and do a lot more:

    >>> SCHEMA = {
    ...     'name': 'name',
    ...     'x': 'x',
    ...     'zed': 'z',
    ...     'p': [{
    ...         '_container': tuple,
    ...         'name': 'name',
    ...         'stuff': lambda p: 42,
    ...         'p': [{
    ...             '_container': tuple,
    ...             'name': 'name'
    ...         }]
    ...     }]
    ... }
    ...
    >>> def use_container_defined_in_schema(obj, schema, depth):
    ...    return schema.get('_container', list)
    ...
    >>> def ignore_private_schema_keys(obj, schema, depth):
    ...     return obj, {k: v for k, v in list(schema.items()) if k[0] != '_'}
    ...
    >>> asserts.dict_equal(
    ...     object_to_dict(
    ...         Point('p1', 5, 2, (
    ...             Point('p2', 3, 4, []),
    ...             Point('p3', 7, 0, [
    ...                 Point('p4', 0, 0, None),
    ...                 Point('p5', 0, 0, None)
    ...             ])
    ...         )),
    ...         schema=SCHEMA,
    ...         callback=ignore_private_schema_keys,
    ...         iterable_callback=use_container_defined_in_schema),
    ...     {
    ...         'name': 'p1', 'x': 5, 'zed': 3,
    ...         'p': (
    ...             {'name': 'p2', 'stuff': 42, 'p': ()},
    ...             {'name': 'p3', 'stuff': 42, 'p': (
    ...                 {'name': 'p4'},
    ...                 {'name': 'p5'}
    ...             )}
    ...         )
    ...     })

    This serializer can optionally adapt schema to objects:

    >>> SCHEMA = [{
    ...     'n': 'name',
    ...     'p': {
    ...         'n': 'name',
    ...         'p': {
    ...             'n': 'name'
    ...         }
    ...     }
    ... }]
    ...
    >>> seen = set()
    >>>
    >>> def reduce_seen(obj, schema, depth):
    ...     if obj in seen:
    ...         return obj, {'r': 'name'}
    ...     seen.add(obj)
    ...     return obj, schema
    ...
    >>> p1 = Point('p1', 0, 0, None)
    >>>
    >>> asserts.list_equal(
    ...     object_to_dict([
    ...         p1,
    ...         Point('p2', 5, 2, Point('p3', 3, 4, p1)),
    ...         Point('p4', 5, 2, Point('p5', 3, 4, p1)),
    ...         Point('p6', 5, 2, p1)
    ...     ], SCHEMA, callback=reduce_seen),
    ...     [
    ...         {'n': 'p1', 'p': None},
    ...         {'n': 'p2', 'p': {'n': 'p3', 'p': {'r': 'p1'}}},
    ...         {'n': 'p4', 'p': {'n': 'p5', 'p': {'r': 'p1'}}},
    ...         {'n': 'p6', 'p': {'r': 'p1'}}
    ...     ])
    """

    if isinstance(schema, list):
        count = len(schema)
        if count != 1:
            raise NotImplementedError(f'List containing {count} items.')
        schema = schema[0]
        container_type = iterable_callback(obj, schema, depth)
        return container_type(
            _object_to_dict_item(i, schema, depth, callback, iterable_callback)
            for i in obj
        )
    return _object_to_dict_item(obj, schema, depth, callback, iterable_callback)


def _object_to_dict_item(
    obj, schema, depth=0,
    callback=lambda o, s, d: (o, s),
    iterable_callback=lambda o, s, d: list
):
    if obj is None:
        return None

    obj_dict = {}
    obj, schema = callback(obj, schema, depth)

    for key, value in list(schema.items()):

        # Direct access to object
        if isinstance(value, str):
            obj_dict[key] = getattr(obj, value)
        elif callable(value):
            obj_dict[key] = value(obj)

        # Nested object(s)
        elif isinstance(value, (dict, list)):
            obj_dict[key] = object_to_dict(
                getattr(obj, key),
                schema[key],
                depth + 1,
                callback,
                iterable_callback)
        else:
            raise NotImplementedError(f'Key {repr(key)} with value {repr(value)}')

    return obj_dict


def dict_to_object(cls, the_dict, inspect_constructor):
    """
    Convert a python dictionary to an instance of a class.

    Set `inspect_constructor` to True to filter input dictionary to avoid sending unexpected keyword
    arguments to the constructor (`__init__`) of `cls`.

    **Example usage**

    >>> from pytoolbox.unittest import asserts
    >>>
    >>> class User(object):
    ...
    ...     def __init__(self, first_name, last_name='Fischer'):
    ...         self.first_name, self.last_name = first_name, last_name
    ...
    ...     @property
    ...     def name(self):
    ...        return '{0} {1}'.format(self.first_name, self.last_name)
    ...
    >>> user_dict = {'first_name': 'Victor', 'last_name': 'Fischer', 'unexpected': 10}
    >>> dict_to_object(User, user_dict, inspect_constructor=False)
    Traceback (most recent call last):
        ...
    TypeError: ...__init__() got an unexpected keyword argument 'unexpected'
    >>> asserts.dict_equal(
    ...     dict_to_object(User, user_dict, inspect_constructor=True).__dict__,
    ...     {'first_name': 'Victor', 'last_name': 'Fischer'})
    """
    if inspect_constructor:
        the_dict = {
            p.name: the_dict.get(p.name, None)
            for p in inspect.signature(cls.__init__).parameters.values()
            if p.name != 'self'
        }
    return cls(**the_dict)


class SlotsToDictMixin(object):

    extra_slots = None

    def to_dict(self, extra_slots=True):
        self_dict = {}
        slots = set(s for s in get_slots(self) if s[0] != '_')
        if extra_slots:
            slots.update(self.__class__.extra_slots or [])
        for attribute in slots:
            value = getattr(self, attribute)
            if value is not None:
                self_dict[attribute] = value
        return self_dict


__all__ = _all.diff(globals())



import abc, re

__all__ = ['Equipement']


class Equipement(object, metaclass=abc.ABCMeta):

    def __init__(self, metadata):
        self.metadata = metadata

    def __bool__(self):
        return bool(self.model)

    def __eq__(self, other):
        try:
            return self.brand == other.brand and self.model == other.model
        except AttributeError:
            return NotImplemented

    def __hash__(self):
        return hash(repr(self))

    def __repr__(self):
        return f'<{self.__class__.__name__} {self.brand} {self.model}>'

    @property
    @abc.abstractmethod
    def brand(self):
        pass

    @property
    def model(self):
        if self.brand and self._model:
            return re.sub(fr'{self.brand}\s+', '', self._model, 1, re.IGNORECASE)
        return self._model

    @property
    @abc.abstractmethod
    def _model(self):
        pass

    @property
    @abc.abstractmethod
    def tags(self):
        pass

    def refresh(self):
        self.__dict__.pop('tags', None)



"""
Module related to enumeration.
"""

import enum


class OrderedEnum(enum.Enum):
    """
    An enumeration both hash-able and ordered by value.

    Reference: https://docs.python.org/3/library/enum.html#orderedenum.
    """

    def __hash__(self):
        return hash(self.value)

    def __eq__(self, other):
        if self.__class__ is other.__class__:
            return self.value == other.value  # pylint:disable=comparison-with-callable
        return NotImplemented

    def __ge__(self, other):
        if self.__class__ is other.__class__:
            return self.value >= other.value  # pylint:disable=comparison-with-callable
        return NotImplemented

    def __gt__(self, other):
        if self.__class__ is other.__class__:
            return self.value > other.value  # pylint:disable=comparison-with-callable
        return NotImplemented

    def __le__(self, other):
        if self.__class__ is other.__class__:
            return self.value <= other.value  # pylint:disable=comparison-with-callable
        return NotImplemented

    def __lt__(self, other):
        if self.__class__ is other.__class__:
            return self.value < other.value  # pylint:disable=comparison-with-callable
        return NotImplemented

    def __ne__(self, other):
        if self.__class__ is other.__class__:
            return self.value != other.value  # pylint:disable=comparison-with-callable
        return NotImplemented



import errno, fcntl, grp, logging, multiprocessing, os, pwd, random, re
import setuptools.archive_util, shlex, shutil, subprocess, threading, time

from . import filesystem, module

_all = module.All(globals())

EMPTY_CMD_RETURN = {'process': None, 'stdout': None, 'stderr': None, 'returncode': None}

# import Popen on steroids if available
try:
    from psutil import NoSuchProcess, Popen
except ImportError:
    from subprocess import Popen
    NoSuchProcess = None

try:
    from shlex import quote
except ImportError:
    from pipes import quote


def kill(process):
    try:
        process.kill()
    except OSError as e:
        if e.errno != errno.ESRCH:
            raise
    except Exception as e:  # pylint:disable=broad-except
        if not NoSuchProcess or not isinstance(e, NoSuchProcess):
            raise


def su(user, group):  # pylint:disable=invalid-name
    """
    Return a function to change current user/group id.

    **Example usage**

    >> import subprocess
    >> subprocess.call(['ls', '/'], preexec_fn=su(1000, 1000))
    >> subprocess.call(['ls', '/'], preexec_fn=su('root', 'root'))
    """
    def set_ids():
        os.setgid(grp.getgrnam(group).gr_gid if isinstance(group, str) else group)
        os.setuid(pwd.getpwnam(user).pw_uid if isinstance(user, str) else user)
    return set_ids


# http://stackoverflow.com/a/7730201/190597
def make_async(fd):  # pylint:disable=invalid-name
    """Add the O_NONBLOCK flag to a file descriptor."""
    fcntl.fcntl(fd, fcntl.F_SETFL, fcntl.fcntl(fd, fcntl.F_GETFL) | os.O_NONBLOCK)


# http://stackoverflow.com/a/7730201/190597
def read_async(fd):  # pylint:disable=invalid-name
    """Read some data from a file descriptor, ignoring EAGAIN errors."""
    try:
        return fd.read()
    except IOError as e:
        if e.errno == errno.EAGAIN:
            return ''
        raise


def to_args_list(args):
    if not args:
        return []
    return shlex.split(args) if isinstance(args, str) else [str(a) for a in args]


def to_args_string(args):
    if not args:
        return ''
    return args if isinstance(args, str) else ' '.join(quote(str(a)) for a in args)


# --------------------------------------------------------------------------------------------------

def raw_cmd(arguments, shell=False, **kwargs):
    """
    Launch a subprocess.

    This function ensure that:

    * subprocess arguments will be converted to a string if `shell` is True
    * subprocess.args is set to the arguments of the subprocess
    """
    arguments = to_args_list(arguments)
    process = Popen(to_args_string(arguments) if shell else arguments, shell=shell, **kwargs)
    if not hasattr(process, 'args'):
        process.args = arguments
    return process


# thanks http://stackoverflow.com/questions/1191374$
def _communicate_with_timeout(*, data, process, input):  # pylint:disable=redefined-builtin
    data['stdout'], data['stderr'] = process.communicate(input=input)


def cmd(  # pylint:disable=too-many-branches,too-many-locals,too-many-statements
    command,
    user=None,
    input=None,  # pylint:disable=redefined-builtin
    cli_input=None,
    cli_output=False,
    communicate=True,
    timeout=None,
    fail=True,
    log=None,
    tries=1,
    delay_min=5,
    delay_max=10,
    **kwargs
):
    """
    Calls the `command` and returns a dictionary with process, stdout, stderr, and the returncode.

    Returned returncode, stdout and stderr will be None if `communicate` is set to False.

    :param user: If set, this will use ``sudo -u <user> ...`` to execute `command` as `user`.
    :type user: unicode
    :param input: If set, sended to stdin (if `communicate` is True).
    :type input: unicode
    :param cli_input: If set, sended to stdin (no condition).
    :type cli_input: unicode
    :param cli_output: Set to True to output (in real-time) stdout to stdout and stderr to stderr.
    :type cli_output: bool
    :param fail: Set to False to avoid the exception `subprocess.CalledProcessError`.
    :type fail: bool
    :param log: A function to log/print details about what is executed/any failure, can be a logger.
    :type log: callable, logging.Logger
    :param communicate: Set to True to communicate with the process, this is a locking call
                        (if timeout is None).
    :type communicate: bool
    :param timeout: Time-out for the communication with the process, in seconds.
    :type timeout: float
    :param tries: How many times you want the command to be retried ?
    :type tries: int
    :param delay_min: Minimum delay to sleep after every attempt communicate must be True.
    :type delay: float, int
    :param delay_max: Maximum delay to sleep after every attempt communicate must be True.
    :type delay: float, int

    * Delay will be a random number in range (`delay_min`, `delay_max`)
    * Set kwargs with any argument of the :mod:`subprocess`.Popen constructor excepting
      stdin, stdout and stderr.

    """

    # convert log argument to logging functions
    log_debug = log_warning = log_exception = None
    if isinstance(log, logging.Logger):
        log_debug, log_warning, log_exception = log.debug, log.warning, log.exception
    elif hasattr(log, '__call__'):
        log_debug = log_warning = log_exception = log

    # create a list and a string of the arguments
    if isinstance(command, str):
        if user is not None:
            command = f'sudo -u {user} {command}'
        args_list, args_string = shlex.split(command), command
    else:
        if user is not None:
            command = ['sudo', '-u', user] + command
        args_list = [str(a) for a in command if a is not None]
        args_string = ' '.join([str(a) for a in command if a is not None])

    # log the execution
    if log_debug:
        log_debug(''.join([
            'Execute ',
            '' if input is None else f'echo {repr(input)} | ',
            args_string,
            '' if cli_input is None else f' < {repr(cli_input)}'
        ]))

    for trial in range(tries):  # noqa
        # create the sub-process
        try:
            process = Popen(
                args_list,
                stdin=subprocess.PIPE,
                stdout=None if cli_output else subprocess.PIPE,
                stderr=None if cli_output else subprocess.PIPE, **kwargs)
        except OSError as e:
            # unable to execute the program (e.g. does not exist)
            if log_exception:
                log_exception(e)
            if fail:
                raise
            return {'process': None, 'stdout': '', 'stderr': e, 'returncode': 2}

        # write to stdin (answer to questions, ...)
        if cli_input is not None:
            process.stdin.write(cli_input)
            process.stdin.flush()

        # interact with the process and wait for the process to terminate
        if communicate:
            data = {}

            thread = threading.Thread(
                target=_communicate_with_timeout,
                kwargs={'data': data, 'input': input, 'process': process})
            thread.start()
            thread.join(timeout=timeout)
            if thread.is_alive():
                try:
                    process.terminate()
                    thread.join()
                except OSError as e:
                    # Manage race condition with process that may terminate just after the call to
                    # thread.is_alive() !
                    if e.errno != errno.ESRCH:
                        raise
            stdout, stderr = data['stdout'], data['stderr']
        else:
            # get a return code that may be None of course ...
            process.poll()
            stdout = stderr = None

        result = {
            'process': process,
            'stdout': stdout,
            'stderr': stderr,
            'returncode': process.returncode
        }

        if process.returncode == 0:
            break

        # failed attempt, may retry
        do_retry = trial < tries - 1
        delay = random.uniform(delay_min, delay_max)
        if log_warning:
            log_warning(' '.join([
                f'Attempt {trial + 1} out of {tries}:',
                f'Will retry in {delay} seconds' if do_retry else 'Failed'
            ]))

        # raise if this is the last try
        if fail and not do_retry:
            raise subprocess.CalledProcessError(process.returncode, args_string, stderr)

        if do_retry:
            time.sleep(delay)

    return result


# --------------------------------------------------------------------------------------------------

def git_add_submodule(directory, url=None, remote='origin', fail=True, log=None, **kwargs):
    if url is not None:
        with open(os.path.join(directory, '.git', 'config'), encoding='utf-8') as f:
            config = f.read()
        regex = rf'\[remote "{remote}"\][^\[]+url\s+=\s+(\S+)'
        url = re.search(regex, config, re.MULTILINE).group(1)
    return cmd(['git', 'submodule', 'add', '-f', url, directory], fail=fail, log=log, **kwargs)


def git_clone_or_pull(directory, url, bare=False, clone_depth=None, reset=True, fail=True, log=None,
                      **kwargs):
    if os.path.exists(directory):
        if reset and not bare:
            cmd(['git', 'reset', '--hard'], cwd=directory, fail=fail, log=log, **kwargs)
        cmd(['git', 'fetch' if bare else 'pull'], cwd=directory, fail=fail, log=log, **kwargs)
    else:
        command = ['git', 'clone']
        if bare:
            command.append('--bare')
        if clone_depth:
            command.extend(['--depth', clone_depth])
        command.extend([url, directory])
        cmd(command, fail=fail, log=log, **kwargs)


# --------------------------------------------------------------------------------------------------

def make(
    archive,
    path=None,
    with_cmake=False,
    configure_options='',
    install=True,
    remove_temporary=True,
    make_options=f'-j{multiprocessing.cpu_count()}',
    fail=True,
    log=None,
    **kwargs
):
    results = {}
    here = os.getcwd()
    path = path or archive.split('.')[0]
    shutil.rmtree(path, ignore_errors=True)
    setuptools.archive_util.unpack_archive(archive, path)
    os.chdir(path)
    if with_cmake:
        filesystem.makedirs('build')
        os.chdir('build')
        results['cmake'] = cmd('cmake -DCMAKE_BUILD_TYPE=RELEASE ..', fail=fail, log=log, **kwargs)
    else:
        results['configure'] = cmd(
            f'./configure {configure_options}', fail=fail, log=log, **kwargs)
    results['make'] = cmd(f'make {make_options}', fail=fail, log=log, **kwargs)
    if install:
        results['make install'] = cmd('make install', fail=fail, log=log, **kwargs)
    os.chdir(here)
    if remove_temporary:
        shutil.rmtree(path)
    return results


# --------------------------------------------------------------------------------------------------

def rsync(  # pylint:disable=too-many-locals
    source,
    destination,
    source_is_dir=False,
    destination_is_dir=False,
    makedest=False,
    archive=True,
    delete=False,
    exclude_vcs=False,
    progress=False,
    recursive=False,
    simulate=False,
    excludes=None,
    includes=None,
    rsync_path=None,
    size_only=False,
    extra=None,
    extra_args=None,
    fail=True,
    log=None,
    **kwargs
):
    if makedest and not os.path.exists(destination):
        # FIXME if dest = remote -> ssh to make dest else make dest
        if extra is None or 'ssh' not in extra:
            os.makedirs(destination)

    source = os.path.normpath(source)
    if os.path.isdir(source) or source_is_dir:
        source = source + os.sep

    destination = os.path.normpath(destination)
    if os.path.isdir(destination) or destination_is_dir:
        destination = destination + os.sep

    command = [
        'rsync',
        '-a' if archive else None,
        '--delete' if delete else None,
        '--progress' if progress else None,
        '-r' if recursive else None,
        '--dry-run' if simulate else None,
        '--size-only' if size_only else None
    ]

    if rsync_path is not None:
        command += ['--rsync-path', rsync_path]
    if extra is not None:
        command += ['-e', extra]
    if excludes is not None:
        command += [f'--exclude={e}' for e in excludes]
    if includes is not None:
        command += [f'--include={i}' for i in includes]
    if exclude_vcs:
        command += ['--exclude=.svn', '--exclude=.git']
    if extra_args is not None:
        command += extra_args
    command += [source, destination]

    return cmd([c for c in command if c], fail=fail, log=log, **kwargs)


def screen_kill(name=None, fail=True, log=None, **kwargs):
    """Kill all screen instances called `name` or all if `name` is None."""
    for instance_name in screen_list(name=name, log=log):
        cmd(['screen', '-S', instance_name, '-X', 'quit'], fail=fail, log=log, **kwargs)


def screen_launch(name, command, fail=True, log=None, **kwargs):
    """Launch a new named screen instance."""
    return cmd(
        ['screen', '-dmS', name] + (command if isinstance(command, list) else [command]),
        fail=fail,
        log=log,
        **kwargs)


def screen_list(name=None, log=None, **kwargs):
    """Returns a list containing all instances of screen. Can be filtered by `name`."""
    screens = cmd(['screen', '-ls', name], fail=False, log=log, **kwargs)['stdout']
    return re.findall(r'\s+(\d+.\S+)\s+\(.*\).*', screens.decode('utf-8'))


def ssh(host, identity_file=None, remote_cmd=None, fail=True, log=None, **kwargs):
    command = ['ssh']
    if identity_file is not None:
        command += ['-i', identity_file]
    command += [host]
    if remote_cmd is not None:
        command += ['-n', remote_cmd]
    return cmd([c for c in command if c], fail=fail, log=log, **kwargs)


__all__ = _all.diff(globals())



import collections, os

from pytoolbox.network.ip import IPSocket
from pytoolbox.network.rtp import RtpPacket
from .base import FecPacket

__all__ = ['FecReceiver']


class FecReceiver(object):  # pylint:disable=too-many-instance-attributes
    """
    A SMPTE 2022-1 FEC streams receiver.
    This receiver accept incoming RTP media and FEC packets and make available the recovered media
    stream.

    **Example usage (with a network capture)**

    ::

        >> import dpkt
        >> from pytoolbox.network.smpte2022.base import FecPacket, RtpPacket
        >> from pytoolbox.network.smpte2022.receiver import FecReceiver
        >> from struct import *
        >>
        >> fec_receiver = FecReceiver(open('test.ts', 'wb'))
        >> fec_receiver.set_delay(1024, FecReceiver.PACKETS)
        >>
        >> for ts, buf in dpkt.pcap.Reader(open('test.dump')):
        >>     udp = dpkt.ethernet.Ethernet(buf).data.data
        >>     if udp.dport == 3300:
        >>         media_packet = RtpPacket.create(
        ..             unpack('!h', udp.data[2:4])[0], unpack('!i', udp.data[4:8])[0],
        ..             RtpPacket.MP2T_PT, bytearray(udp.data[12:]))
        >>         fec_receiver.put_media(media_packet, onlyMP2TS=True)
        >>     elif udp.dport in (3302, 3304):
        >>         fec_data = bytearray(udp.data)
        >>         fec_receiver.put_fec(FecPacket(fec_data, len(fec_data)))
        >> print(fec_receiver)
        >> fec_receiver.flush()

    **Example usage**

    Media packets are sorted by the buffer, so, it's time to test this feature:

    >>> import io, random
    >>> output = io.StringIO()
    >>> receiver = FecReceiver(output)
    >>> receiver.set_delay(1000, FecReceiver.PACKETS)
    >>> source = list(range(30))
    >>> random.shuffle(source)
    >>> for i in source:
    ...     receiver.put_media(RtpPacket.create(i, i * 100, RtpPacket.MP2T_PT, str(i)), True)
    >>> receiver.flush()
    >>> output.getvalue()
    '01234567891011121314151617181920212223242526272829'

    Testing FEC algorithm correctness:

    >>> import io, os, random
    >>> output = io.BytesIO()
    >>> receiver = FecReceiver(output)
    >>> receiver.set_delay(1024, FecReceiver.PACKETS)
    >>> L, D = 4, 5
    >>> # Generate a [D][L] matrix of randomly generated RTP packets
    >>> matrix = [[RtpPacket.create(L * j + i, (L * j + i) * 100 + random.randint(0, 50),
    ...           RtpPacket.MP2T_PT, bytearray(os.urandom(random.randint(50, 100))))
    ...           for i in range(L)] for j in range(D)]
    >>> assert len(matrix) == D and len(matrix[0]) == L
    >>> # Retrieve the first column of the matrix
    >>> for column in matrix:
    ...     for media in column:
    ...         if media.sequence != 0:
    ...             receiver.put_media(media, True)
    >>> fec = FecPacket.compute(1, FecPacket.XOR, FecPacket.COL, L, D, [p[0] for p in matrix[0:]])
    >>> print('dir={0} snbase={1} offset={2} na={3}'.format(
    ...     fec.direction, fec.snbase, fec.offset, fec.na))
    dir=0 snbase=0 offset=4 na=5
    >>> receiver.put_fec(fec)
    >>> print(receiver)
    Name  Received Buffered Maximum Dropped
    Media       19       20      20
    Col          1        0       1       0
    Row          0        0       0       0
    Cross                 0       1
    FEC statistics, media packets :
    Recovered Aborted Overwritten Missing
            1       0           0       0
    Current position (media sequence) : 0
    Current delay (can be set) : 20 packets
    FEC matrix size (LxD) : 4x5 = 20 packets
    >>> receiver.flush()

    The output does begin by recovered first packet of the matrix:

    >>> print(matrix[0][1].payload == output.getvalue()[:len(matrix[0][1].payload)])
    False
    >>> print(matrix[0][0].payload == output.getvalue()[:len(matrix[0][0].payload)])
    True
    """

    # <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< Constants >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

    ER_DELAY_UNITS = "Unknown delay units '{0}'"
    ER_DIRECTION = 'FEC packet direction is neither COL nor ROW : {0}'
    ER_FLUSHING = 'Currently flushing buffers'
    ER_MISSING_COUNT = 'They are {0} missing media packet, expected one (1)'
    ER_FEC_DIRECTION = "Invalid FEC direction '{0}'"
    ER_COL_MISMATCH = 'Column FEC packet n°{0}, expected n°{1}'
    ER_COL_OVERWRITE = \
        'Another column FEC packet is already registered to protect media packet n°{0}'
    ER_ROW_MISMATCH = 'Row FEC packet n°{0}, expected n°{1}'
    ER_ROW_OVERWRITE = 'Another row FEC packet is already registered to protect media packet n°{0}'
    ER_GET_COL_CASCADE = \
        'Column FEC cascade : Unable to compute sequence # of the media packet to recover{0}{1}{0}'
    ER_GET_ROW_CASCADE = \
        'Row FEC cascade : Unable to compute sequence # of the media packet to recover{0}{1}{0}'
    ER_NULL_COL_CASCADE = 'Column FEC cascade : Unable to find linked entry in crosses buffer'
    ER_NULL_ROW_CASCADE = 'Row FEC cascade : Unable to find linked entry in crosses buffer'
    ER_STARTUP = 'Current position still not initialized (startup state)'
    ER_VALID_RTP_MP2TS = 'packet is not valid (expected RTP packet + MPEG2-TS payload)'
    ER_VALID_RTP = 'packet is not valid (expected RTP packet)'

    DELAY_NAMES = ['packets', 'seconds']
    DELAY_RANGE = range(len(DELAY_NAMES))  # noqa
    PACKETS, SECONDS = DELAY_RANGE

    # <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< Constructors >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

    def __init__(self, output):
        """
        Construct a new FecReceiver and register `output`.

        :param output: Where to output payload of the recovered stream.
        :type output: IOBase

        **Example usage**

        Not yet an output:

        >>> FecReceiver(None)
        Traceback (most recent call last):
            ...
        ValueError: output is None
        >>>
        >>> import io
        >>> output = io.StringIO()
        >>> receiver = FecReceiver(output)
        """
        if not output:
            raise ValueError('output is None')
        # Media packets storage, medias[media seq] = media pkt
        self.medias = {}
        self.startup = True    # Indicate that actual position must be initialized
        self.flushing = False  # Indicate that a flush operation is actually running
        self.position = 0      # Actual position (sequence number) in the medias buffer
        # Link media packets to fec packets able to recover it, crosses[mediaseq] = {colseq, rowseq}
        self.crosses = {}
        # Fec packets + related information storage, col[sequence] = { fec pkt + info }
        self.cols = {}
        self.rows = {}
        self.matrixL = 0  # Detected FEC matrix size (number of columns) pylint:disable=invalid-name
        self.matrixD = 0  # Detected FEC matrix size (number of rows)    pylint:disable=invalid-name
        # Output
        self.output = output  # Registered output
        # Settings
        self.delay_value = 100           # RTP buffer delay value
        self.delay_units = self.PACKETS  # RTP buffer delay units
        # Statistics about media (buffers and packets)
        self.media_received = 0          # Received media packets counter
        self.media_recovered = 0         # Recovered media packets counter
        self.media_aborted_recovery = 0  # Aborted media packet recovery counter
        self.media_overwritten = 0       # Overwritten media packets counter
        self.media_missing = 0           # Missing media packets counter
        self.max_media = 0               # Largest amount of stored elements in the medias buffer
        # Statistics about fec (buffers and packets)
        self.col_received = 0  # Received column fec packets counter
        self.row_received = 0  # Received row fec packets counter
        self.col_dropped = 0   # Dropped column fec packets counter
        self.row_dropped = 0   # Dropped row fec packets counter
        self.max_cross = 0     # Largest amount of stored elements in the crosses buffer
        self.max_col = 0       # Largest amount of stored elements in the columns buffer
        self.max_row = 0       # Largest amount of stored elements in the rows buffer
        self.lostogram = collections.defaultdict(int)  # Statistics about lost medias
        self.lostogram_counter = 0  # Incremented while there are lost media packets

    # <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< Properties >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

    @property
    def current_delay(self):
        """Return current delay based on the length of the media buffer."""
        if len(self.medias) == 0:
            return 0
        if self.delay_units == self.PACKETS:
            return len(self.medias)
        if self.delay_units == self.SECONDS:
            raise NotImplementedError()
        raise ValueError(self.ER_DELAY_UNITS.format(self.delay_units))
        # return medias.lastEntry().getValue().getTime() - medias.firstEntry().getValue().getTime()

    # <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< Functions >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

    def set_delay(self, value, units):
        """Set desired size for the internal media buffer."""
        if units == self.PACKETS:
            self.delay_value = value
            self.delay_units = units
        elif units == self.SECONDS:
            raise NotImplementedError()
        else:
            raise ValueError(self.ER_DELAY_UNITS.format(units))

    def put_media(self, media, onlyMP2TS):  # pylint:disable=invalid-name
        """Put an incoming media packet."""
        if self.flushing:
            raise ValueError(self.ER_FLUSHING)
        if onlyMP2TS:
            if not media.validMP2T:
                raise ValueError(self.ER_VALID_RTP_MP2TS)
        else:
            if not media.valid:
                raise ValueError(self.ER_VALID_RTP)

        # Put the media packet into medias buffer
        if media.sequence in self.medias:
            self.media_overwritten += 1
        self.medias[media.sequence] = media
        if len(self.medias) > self.max_media:
            self.max_media = len(self.medias)
        self.media_received += 1

        cross = self.crosses.get(media.sequence)
        if cross:
            # Simulate the recovery of a media packet to update buffers and potentially start
            self.recover_media_packet(media.sequence, cross, None)  # a recovery cascade !

        self.out()  # FIXME maybe better to call it from another thread

    def put_fec(self, fec):  # pylint:disable=too-many-branches,too-many-statements
        """
        Put an incoming FEC packet, the algorithm will do the following according to these
        scenarios:

        1. The fec packet is useless if none of the protected media packets is missing
        2. Only on media packet missing, fec packet is able to recover it now !
        3. More than one media packet is missing, fec packet stored for future recovery
        """
        if self.flushing:
            raise ValueError(self.ER_FLUSHING)
        if not fec.valid:
            raise ValueError('Invalid FEC packet')

        if fec.direction == FecPacket.COL:
            self.col_received += 1
        elif fec.direction == FecPacket.ROW:
            self.row_received += 1
        else:
            raise ValueError(self.ER_DIRECTION.format(fec.direction))

        cross = None
        media_lost = 0
        media_max = (fec.snbase + fec.na * fec.offset) & RtpPacket.S_MASK
        media_test = fec.snbase

        while media_test != media_max:
            # If media packet is not in the medias buffer (is missing)
            if media_test not in self.medias:
                media_lost = media_test
                # TODO
                cross = self.crosses.get(media_test)
                if not cross:
                    cross = {'col_sequence': None, 'row_sequence': None}
                    self.crosses[media_test] = cross
                    if len(self.crosses) > self.max_cross:
                        self.max_cross = len(self.crosses)

                # Register the fec packet able to recover the missing media packet
                if fec.direction == FecPacket.COL:
                    if cross['col_sequence']:
                        raise ValueError(self.ER_COL_OVERWRITE.format(media_lost))
                    cross['col_sequence'] = fec.sequence

                elif fec.direction == FecPacket.ROW:
                    if cross['row_sequence']:
                        raise ValueError(self.ER_ROW_OVERWRITE.format(media_lost))
                    cross['row_sequence'] = fec.sequence

                else:
                    raise ValueError(self.ER_FEC_DIRECTION.format(fec.direction))

                fec.set_missing(media_test)

            media_test = (media_test + fec.offset) & RtpPacket.S_MASK

        if fec.L:
            self.matrixL = fec.L

        if fec.D:
            self.matrixD = fec.D

        # [1] The fec packet is useless if none of the protected media packets is missing
        if len(fec.missing) == 0:
            return

        # FIXME check if 10*delay_value is a good way to avoid removing early fec packets !
        # The fec packet is useless if it needs an already output'ed media packet to do recovery
        drop = not self.validity_window(
            fec.snbase,
            self.position,
            (self.position + 10 * self.delay_value) & RtpPacket.S_MASK
        )

        if fec.direction == FecPacket.COL:
            if drop:
                self.col_dropped += 1
                return
            self.cols[fec.sequence] = fec
            if len(self.cols) > self.max_col:
                self.max_col = len(self.cols)

        if fec.direction == FecPacket.ROW:
            if drop:
                self.row_dropped += 1
                return
            self.rows[fec.sequence] = fec
            if len(self.rows) > self.max_row:
                self.max_row = len(self.rows)

        # [2] Only on media packet missing, fec packet is able to recover it now !
        if len(fec.missing) == 1:
            self.recover_media_packet(media_lost, cross, fec)
            self.out()  # FIXME maybe better to call it from another thread
        # [3] More than one media packet is missing, fec packet stored for future recovery

    def flush(self):
        """Flush all buffers and output media packets to registered output (``self.output``)."""
        try:
            self.flushing = True
            self.out()
            self.output.flush()
        finally:
            self.flushing = False

    def cleanup(self):
        """Remove FEC packets that are stored / waiting but useless."""
        if self.flushing:
            raise ValueError(self.ER_FLUSHING)

        if self.startup:
            raise ValueError(self.ER_STARTUP)

        if self.delay_units == self.PACKETS:
            start, end = self.position, (self.position + self.delay_value) & RtpPacket.S_MASK
            for media_sequence, cross in list(self.crosses.items()):
                if not self.validity_window(media_sequence, start, end):
                    del self.cols[cross['col_sequence']]
                    del self.rows[cross['row_sequence']]
                    del self.crosses[media_sequence]
        elif self.delay_units == self.SECONDS:
            raise NotImplementedError()

        raise ValueError(self.ER_DELAY_UNITS.format(self.delay_units))

    def recover_media_packet(
        self,
        media_sequence,
        cross,
        fec
    ):  # pylint:disable=too-many-branches,too-many-locals,too-many-statements
        """
        Recover a missing media packet helped by a FEC packet, this method is also called to
        register an incoming media packet if it is registered as missing.
        """
        recovered_by_fec = fec is not None

        # Read and remove "cross" it from the buffer
        col_sequence = cross['col_sequence']
        row_sequence = cross['row_sequence']
        del self.crosses[media_sequence]

        # Recover the missing media packet and remove any useless linked fec packet
        if recovered_by_fec:

            if len(fec.missing) != 1:
                raise NotImplementedError(self.ER_MISSING_COUNT.format(len(fec.missing)))

            if fec.direction == FecPacket.COL and fec.sequence != col_sequence:
                raise NotImplementedError(self.ER_COL_MISMATCH.format(fec.sequence, col_sequence))

            if fec.direction == FecPacket.ROW and fec.sequence != row_sequence:
                raise NotImplementedError(self.ER_ROW_MISMATCH.format(fec.sequence, row_sequence))

            # Media packet recovery
            # > Copy fec packet fields into the media packet
            media = RtpPacket.create(
                media_sequence,
                fec.timestamp_recovery,
                fec.payload_type_recovery,
                fec.payload_recovery)
            payload_size = fec.length_recovery

            # > recovered payload ^= all media packets linked to the fec packet
            aborted = False
            media_max = (fec.snbase + fec.na * fec.offset) & RtpPacket.S_MASK
            media_test = fec.snbase
            while media_test != media_max:
                if media_test == media_sequence:
                    media_test = (media_test + fec.offset) & RtpPacket.S_MASK
                    continue
                friend = self.medias[media_test]
                # Unable to recover the media packet if any of the friend media packets is missing
                if not friend:
                    self.media_aborted_recovery += 1
                    aborted = True
                    break
                media.payload_type ^= friend.payload_type
                media.timestamp ^= friend.timestamp
                payload_size ^= friend.payload_size
                # TODO Optimize this
                for index in range(min(len(media.payload), len(friend.payload))):
                    media.payload[index] ^= friend.payload[index]
                media_test = (media_test + fec.offset) & RtpPacket.S_MASK

            # If the media packet is successfully recovered
            if not aborted:
                media.payload = media.payload[0:payload_size]
                self.media_recovered += 1
                if media.sequence in self.medias:
                    self.media_overwritten += 1
                self.medias[media.sequence] = media
                if len(self.medias) > self.max_media:
                    self.max_media = len(self.medias)
                if fec.direction == FecPacket.COL:
                    del self.cols[fec.sequence]
                else:
                    del self.rows[fec.sequence]

        # Check if a cascade effect happens ...
        fec_col = self.cols.get(col_sequence) if col_sequence else None
        fec_row = self.rows.get(row_sequence) if row_sequence else None

        if fec_col:
            fec_col.set_recovered(media_sequence)
        if fec_row:
            fec_row.set_recovered(media_sequence)

        if fec_col:
            if len(fec_col.missing) == 1:
                # Cascade !
                cascade_media_sequence = fec_col.missing[0]
                if cascade_media_sequence:
                    cascade_cross = self.crosses.get(cascade_media_sequence)
                    if cascade_cross:
                        self.recover_media_packet(cascade_media_sequence, cascade_cross, fec_col)
                    else:
                        raise NotImplementedError(
                            f'recover_media_packet({media_sequence}, {cross}, {fec}):{os.linesep}'
                            f'{self.ER_NULL_COL_CASCADE}{os.linesep}'
                            f'media sequence : {cascade_media_sequence}{os.linesep}'
                            f'{fec_col}{os.linesep}')
                else:
                    raise NotImplementedError(self.ER_GET_COL_CASCADE.format(os.linesep, fec_col))

        if fec_row:
            if len(fec_row.missing) == 1:
                # Cascade !
                cascade_media_sequence = fec_row.missing[0]
                if cascade_media_sequence:
                    cascade_cross = self.crosses.get(cascade_media_sequence)
                    if cascade_cross:
                        self.recover_media_packet(cascade_media_sequence, cascade_cross, fec_row)
                    else:
                        raise NotImplementedError(
                            f'{self.ER_NULL_ROW_CASCADE}{os.linesep}'
                            f'recover_media_packet({media_sequence}, {cross}, {fec}):{os.linesep}'
                            f'media sequence : {cascade_media_sequence}{os.linesep}'
                            f'{fec_row}{os.linesep}')
                else:
                    raise NotImplementedError(self.ER_GET_ROW_CASCADE.format(os.linesep, fec_row))

    def out(self):
        """Extract packets to output in order to keep a 'certain' amount of them in the buffer."""
        units = self.PACKETS if self.flushing else self.delay_units
        value = 0 if self.flushing else self.delay_value
        lostogram_counter = 0

        # Extract packets to output in order to keep a 'certain' amount of them in the buffer
        if units == self.PACKETS:  # based on buffer size
            while len(self.medias) > value:
                # Initialize or increment actual position (expected sequence number)
                if self.startup:
                    self.position = min(self.medias.keys()) & RtpPacket.S_MASK
                else:
                    self.position = (self.position + 1) & RtpPacket.S_MASK

                self.startup = False

                media = self.medias.get(self.position)
                if media:
                    self.lostogram[lostogram_counter] += 1
                    lostogram_counter = 0
                    del self.medias[media.sequence]
                    if self.output:
                        self.output.write(media.payload)
                else:
                    self.media_missing += 1
                    lostogram_counter += 1

                # Remove any fec packet linked to current media packet
                cross = self.crosses.get(self.position)
                if cross:
                    del self.crosses[self.position]
                    if cross['col_sequence'] and cross['col_sequence'] in self.cols:
                        del self.cols[cross['col_sequence']]
                    if cross['row_sequence'] and cross['row_sequence'] in self.rows:
                        del self.rows[cross['row_sequence']]

        # Extract packets to output in order to keep a 'certain' amount of them in the buffer
        elif units == self.SECONDS:  # based on time stamps
            raise NotImplementedError()
        else:
            raise ValueError(self.ER_DELAY_UNITS.format(units))

    def __str__(self):
        """
        Return a string representing this instance.

        **Example usage**

        >>> print(FecReceiver('salut'))
        Name  Received Buffered Maximum Dropped
        Media        0        0       0
        Col          0        0       0       0
        Row          0        0       0       0
        Cross                 0       0
        FEC statistics, media packets :
        Recovered Aborted Overwritten Missing
                0       0           0       0
        Current position (media sequence) : 0
        Current delay (can be set) : 0 packets
        FEC matrix size (LxD) : 0x0 = 0 packets
        """
        delay_format = '%.0f' if self.delay_units == self.PACKETS else '%.2f'
        delay_msg = f'{delay_format % self.current_delay} {self.DELAY_NAMES[self.delay_units]}'

        return (  # pylint:disable=consider-using-f-string
            "Name  Received Buffered Maximum Dropped{0}"
            "Media %8d%9d%8d{0}"
            "Col   %8d%9d%8d%8d{0}"
            "Row   %8d%9d%8d%8d{0}"
            "Cross         %9d%8d{0}"
            "FEC statistics, media packets :{0}"
            "Recovered Aborted Overwritten Missing{0}"
            "%9d%8d%12d%8d{0}"
            "Current position (media sequence) : %s{0}"
            "Current delay (can be set) : %s{0}"
            "FEC matrix size (LxD) : %sx%s = %s packets".format(os.linesep) %
            (self.media_received, len(self.medias), self.max_media,
             self.col_received, len(self.cols), self.max_col, self.col_dropped,
             self.row_received, len(self.rows), self.max_row, self.row_dropped,
             len(self.crosses), self.max_cross, self.media_recovered,
             self.media_aborted_recovery, self.media_overwritten, self.media_missing,
             self.position, delay_msg, self.matrixL, self.matrixD, self.matrixL * self.matrixD))

    # <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<< Static >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

    @staticmethod
    def compute_col_address(media_socket):
        """
        Compute column FEC socket based on media stream socket (port +2).

        **Example usage**

        >>> from pytoolbox.unittest import asserts
        >>> asserts.dict_equal(FecReceiver.compute_col_address('192.168.50.100:8000'), {
        ...     'ip': '192.168.50.100', 'port': 8002
        ... })
        >>> asserts.dict_equal(FecReceiver.compute_col_address(IPSocket('50.0.0.7:4000')), {
        ...     'ip': '50.0.0.7', 'port': 4002
        ... })
        >>> print(FecReceiver.compute_col_address('salut'))
        Traceback (most recent call last):
            ....
        pytoolbox.exceptions.InvalidIPSocketError: salut is not a valid IP socket.
        """
        if not isinstance(media_socket, dict):
            media_socket = IPSocket(media_socket)
        media_socket['port'] += 2
        return media_socket

    @staticmethod
    def compute_row_address(media_socket):
        """
        Compute column FEC socket based on media stream socket (port +4).

        **Example usage**

        >>> from pytoolbox.unittest import asserts
        >>> asserts.dict_equal(FecReceiver.compute_row_address('192.168.50.100:8000'), {
        ...     'ip': '192.168.50.100', 'port': 8004
        ... })
        >>> asserts.dict_equal(FecReceiver.compute_row_address(IPSocket('50.0.0.7:4000')), {
        ...     'ip': '50.0.0.7', 'port': 4004
        ... })
        >>> print(FecReceiver.compute_row_address('salut'))
        Traceback (most recent call last):
            ....
        pytoolbox.exceptions.InvalidIPSocketError: salut is not a valid IP socket.
        """
        if not isinstance(media_socket, dict):
            media_socket = IPSocket(media_socket)
        media_socket['port'] += 4
        return media_socket

    @staticmethod
    def validity_window(current, start, end):
        """
        Returns True if `current` is in the validity window bounded by `start` and `end`.

        This method is circular-buffer aware and they are 2 cases (validity window ``[====]``)::

            1) start=     6 end=9 :   0   1  2 3 4 5 [=======] 10 ... 65'533  65'534 65'535
            2) start=65'534 end=1 :  ======] 2 3 4 5  6 7 8 9  10 ... 65'533 [=============

        **Example usage**

        Testing validity window condition:

        >>> FecReceiver.validity_window(0, 5, 10)
        False
        >>> FecReceiver.validity_window(5, 5, 10)
        True
        >>> FecReceiver.validity_window(8, 5, 10)
        True
        >>> FecReceiver.validity_window(10, 5, 10)
        True
        >>> FecReceiver.validity_window(15, 5, 10)
        False
        >>> FecReceiver.validity_window(0, 65534, 2)
        True
        >>> FecReceiver.validity_window(2, 65534, 2)
        True
        >>> FecReceiver.validity_window(5, 65534, 2)
        False
        >>> FecReceiver.validity_window(65534, 65534, 2)
        True
        >>> FecReceiver.validity_window(65535, 65534, 2)
        True
        """
        if end > start:
            return start <= current <= end
        return current <= end or current >= start



import copy, re

from django.core import validators
from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.utils.deconstruct import deconstructible
from django.utils.translation import gettext_lazy as _

__all__ = ['EmptyValidator', 'KeysValidator', 'MD5ChecksumValidator']


class EmptyValidator(validators.RegexValidator):
    regex = r'\S+'
    message = _('This field cannot be blank.')
    code = 'blank'


@deconstructible
class KeysValidator(object):
    """
    A validator designed for HStore to require, even restrict keys.

    Code mostly borrowed from: https://github.com/django/django/blob/master/django/contrib/postgres/validators.py
    """

    messages = {
        'missing_keys': _('Some keys were missing: %(keys)s'),
        'extra_keys': _('Some unknown keys were provided: %(keys)s'),
    }
    strict = False

    def __init__(self, required_keys=None, optional_keys=None, strict=False, messages=None):
        self.required_keys = set(required_keys or [])
        self.optional_keys = set(optional_keys or [])
        if not self.required_keys and not self.optional_keys:
            raise ImproperlyConfigured('You must set at least `required_keys` or `optional_keys`')
        self.strict = strict
        if messages is not None:
            self.messages = copy.copy(self.messages)
            self.messages.update(messages)

    def __call__(self, value):
        keys = set(value.keys())
        if self.required_keys:
            missing_keys = self.required_keys - keys
            if missing_keys:
                raise ValidationError(
                    self.messages['missing_keys'],
                    code='missing_keys',
                    params={'keys': ', '.join(missing_keys)})
        if self.strict:
            extra_keys = keys - self.required_keys - self.optional_keys
            if extra_keys:
                raise ValidationError(
                    self.messages['extra_keys'],
                    code='extra_keys',
                    params={'keys': ', '.join(extra_keys)})

    def __eq__(self, other):
        return (
            isinstance(other, self.__class__) and
            self.required_keys == other.required_keys and
            self.optional_keys == other.optional_keys and
            self.messages == other.messages and
            self.strict == other.strict
        )

    def __ne__(self, other):
        return not self == other


class MD5ChecksumValidator(validators.RegexValidator):
    regex = re.compile(r'[0-9a-f]{32}')



import itertools, os

from .filesystem import find_recursive
from .subprocess import rsync

__all__ = ['relocate']


def relocate(source_directory, destination_directory, encoding='utf-8'):
    """
    Copy and relocate a Python virtual environment.
    Update the paths in `*.egg-link, *.pth, *.pyc`, etc.
    """

    if not os.path.exists(destination_directory):
        rsync(
            source_directory,
            destination_directory,
            destination_is_dir=True,
            makedest=True,
            recursive=True
        )

    b_source_directory = source_directory.encode(encoding)
    b_destination_directory = destination_directory.encode(encoding)

    for path in itertools.chain.from_iterable([
        find_recursive(destination_directory, ['*.egg-link', '*.pth', '*.pyc', 'RECORD']),
        find_recursive(os.path.join(destination_directory, 'bin'), '*'),
        find_recursive(os.path.join(destination_directory, 'src'), '*.so')
    ]):
        with open(path, 'r+b') as f:
            content = f.read().replace(b_source_directory, b_destination_directory)
            f.seek(0)
            f.write(content)
            f.truncate()



"""
Extra `permission policies <http://www.django-rest-framework.org/tutorial/4-authentication-and-permissions/>`_
for building your own `Django REST Framework <https://github.com/tomchristie/django-rest-framework>`_
powered API.
"""

from oauth2_provider.ext.rest_framework import TokenHasReadWriteScope
from rest_framework import permissions

__all__ = ['IsAuthenticatedOrTokenHasReadWriteScope']


class IsAuthenticatedOrTokenHasReadWriteScope(permissions.BasePermission):

    permissions = [permissions.IsAuthenticated(), TokenHasReadWriteScope()]

    def has_permission(self, request, view):
        return any(p.has_permission(request, view) for p in self.permissions)

    def has_object_permission(self, request, view, obj):
        return any(p.has_object_permission(request, view, obj) for p in self.permissions)



"""
Module related to file system and path operations.
"""

import collections, copy, errno, grp, os, pwd, shutil, tempfile, time, uuid

import magic

from . import module
from .datetime import datetime_now
from .regex import from_path_patterns

_all = module.All(globals())


def chown(path, user=None, group=None, recursive=False, **walk_kwargs):
    """
    Change owner/group of a path, can be recursive.
    Both can be a name, an id or None to leave it unchanged.
    """
    uid = to_user_id(user)
    gid = to_group_id(group)
    if recursive:
        for dirpath, _, filenames in os.walk(path, **walk_kwargs):
            os.chown(dirpath, uid, gid)
            for filename in filenames:
                os.chown(os.path.join(dirpath, filename), uid, gid)
    else:
        os.chown(path, uid, gid)


def find_recursive(directory, patterns, regex=False, **walk_kwargs):
    """
    Yield filenames matching any of the patterns.
    Patterns will be compiled to regular expressions, if necessary.

    If `regex` is set to True, then any string pattern will be converted from the unix-style
    wildcard to the regular expression equivalent using :func:`fnatmch.translate`.

    **Example usage**

    >>> import re
    >>> from pathlib import Path
    >>>
    >>> directory = Path(__file__).resolve().parent
    >>>
    >>> next(find_recursive(directory, '*/collections*'))
    '.../pytoolbox/collections.py'
    >>> filenames = sorted(find_recursive(directory, ['*/django*', '*/*.py']))
    >>> [Path(f).name for f in filenames[-4:]]
    ['unittest.py', 'validation.py', 'virtualenv.py', 'voluptuous.py']
    >>> str(directory / 'aws' / 's3.py') in filenames
    True
    >>> str(directory / 'django') in filenames  # Its a directory
    False
    >>> a_files = set(find_recursive(directory, re.compile(r'.*/st.+\\.py$')))
    >>> b_files = set(find_recursive(directory, ['.*/st.+\\.py$'], regex=True))
    >>> a_files == b_files
    True
    >>> [Path(f).name for f in sorted(a_files)]
    ['storage.py', 'states.py', 'string.py']
    """
    patterns = from_path_patterns(patterns, regex=regex)
    for dirpath, _, filenames in os.walk(directory, **walk_kwargs):
        for filename in filenames:
            filename = os.path.join(dirpath, filename)
            if any(p.match(filename) for p in patterns):
                yield filename


def file_mime(path, mime=True):
    """
    Return file mime type.

    **Example usage**

    >>> from pathlib import Path
    >>>
    >>> directory = Path(__file__).resolve().parent
    >>>
    >>> file_mime(directory / '..' / 'setup.cfg')
    'text/plain'
    >>> file_mime(directory / 'filesystem.py') in ('text/plain', 'text/x-python')
    True
    >>> file_mime('missing-file') is None
    True
    """
    try:
        return magic.from_file(str(path), mime=mime)
    except OSError:
        return None


def first_that_exist(*paths):
    """
    Returns the first file/directory that exist.

    **Example usage**

    >>> from pathlib import Path
    >>>
    >>> directory = Path(__file__).resolve().parent
    >>>
    >>> first_that_exist('', '/etc', '.')
    '/etc'
    >>> first_that_exist('', directory) == directory
    True
    >>> first_that_exist('does_not_exist.com', '', '..')
    '..'
    >>> first_that_exist('does_not_exist.ch') is None
    True
    """
    for path in paths:
        if os.path.exists(path):
            return path
    return None


def from_template(
    template,
    destination,
    values,
    is_file=True,
    jinja2=False,
    pre_func=None,
    post_func=None,
    directories='.'
):
    """
    Return a `template` rendered with `values` using string.format or Jinja2 as the template engine.

    * Set `destination` to a filename to store the output, to a Falsy value to skip this feature.
    * Set `is_file` to False to use value of `template` as the content and not a filename to read.
    * Set `{pre,post}_func` to a callback function with the signature f(content, values, jinja2)
    * Set `directories` to the paths where the Jinja2 loader will lookup for *base* templates.

    **Example usage**

    >>> template = '{{username={user}; password={pass}}}'
    >>> values = {'user': 'tabby', 'pass': 'miaow', 'other': 10}
    >>> with open('config.template', 'w', encoding='utf-8') as f:
    ...     f.write(template)
    36

    The behavior when all keys are given (and even more):

    >>> _ = from_template('config.template', 'config', values)
    >>> open('config').read()
    '{username=tabby; password=miaow}'

    >>> _ = from_template(template, 'config', values, is_file=False)
    >>> open('config').read()
    '{username=tabby; password=miaow}'

    >>> def post_func(content, values, jinja2):
    ...     return content.replace('tabby', 'tikky')
    >>> from_template(template, None, values, is_file=False, post_func=post_func)
    '{username=tikky; password=miaow}'

    The behavior if a value for a key is missing:

    >>> from_template('config.template', 'config', {'pass': 'miaow', 'other': 10})
    Traceback (most recent call last):
        ...
    KeyError: ...'user'
    >>> open('config').read()
    '{username=tabby; password=miaow}'

    >>> os.remove('config.template')
    >>> os.remove('config')
    """
    if is_file:
        with open(template, encoding='utf-8') as f:
            content = f.read()
    else:
        content = template
    if pre_func:
        content = pre_func(content, values=values, jinja2=jinja2)
    if jinja2:
        from jinja2 import Environment, FileSystemLoader, StrictUndefined
        loader = FileSystemLoader(directories)
        environment = Environment(loader=loader, undefined=StrictUndefined)
        content = environment.from_string(content).render(**values)
    else:
        content = content.format(**values)
    if post_func:
        content = post_func(content, values=values, jinja2=jinja2)
    if destination:
        with open(destination, 'w', encoding='utf-8') as f:
            f.write(content)
    return content


def get_bytes(path_or_data, encoding='utf-8', is_path=False, chunk_size=None):
    """
    Yield the content read from the given `path` or the `data` converted to bytes.

    Remark: Value of `encoding` is used only if `data` is actually a string.
    """
    if is_path:
        with open(path_or_data, 'rb') as f:
            if chunk_size is not None:
                while True:
                    data = f.read(chunk_size)
                    if not data:
                        break
                    yield data
            else:
                yield f.read()
    else:
        yield path_or_data.encode(encoding) if isinstance(path_or_data, str) else path_or_data


def get_size(path, patterns='*', regex=False, **walk_kwargs):
    """
    Returns the size of a file or directory.

    If given `path` is a directory (or symlink to a directory), then returned value is computed by
    summing the size of all files, and that recursively.

    **Example usage**

    >>> from pathlib import Path
    >>>
    >>> directory = Path(__file__).resolve().parent
    >>>
    >>> get_size(directory / '..' / 'setup.py')
    8102
    >>> get_size(directory/ '..', '*.cfg')
    105
    >>> get_size(directory / '..', '.*/v[^/]+\\.py', regex=True) > 10000
    True
    """
    if os.path.isfile(path):
        return os.stat(path).st_size
    return sum(
        os.stat(f).st_size
        for f in find_recursive(path, patterns, regex=regex, **walk_kwargs))


def makedirs(path, mode=0o777, parent=False):
    """
    Recursively make directories (which may already exists) without throwing an exception.
    Returns True if operation is successful, False if directory found and re-raise any other type
    of exception (raised by `os.makedirs`).

    **Example usage**

    >>> import os, shutil
    >>> from pathlib import Path
    >>>
    >>> filesystem_py = Path(__file__).resolve()
    >>>
    >>> makedirs('/etc')
    False
    >>> makedirs('/tmp/salut/mec')
    True
    >>> shutil.rmtree('/tmp/salut')
    >>>
    >>> makedirs('/tmp/some/path/file.txt', parent=True)
    True
    >>> os.path.exists('/tmp/some/path')
    True
    >>> os.path.exists('/tmp/some/path/file.txt')
    False
    >>> shutil.rmtree('/tmp/some')
    >>>
    >>> makedirs(filesystem_py)
    Traceback (most recent call last):
        ...
    FileExistsError: ...
    """
    if parent:
        path = os.path.dirname(path)
    try:
        os.makedirs(path, mode=mode)
        return True
    except OSError as e:
        # Directory exists
        if e.errno == errno.EEXIST and os.path.isdir(path):
            return False
        raise  # Re-raise exception if a different error occurred


def recursive_copy(  # pylint:disable=too-many-locals
    source_path,
    destination_path,
    progress_callback=None,
    ratio_delta=0.01,
    time_delta=1,
    check_size=True,
    remove_on_error=True
):
    """
    Copy the content of a source directory to a destination directory.
    This function is based on a block-copy algorithm making progress update possible.

    Given `progress_callback` will be called with *start_date*, *elapsed_time*, *eta_time*,
    *src_size*, *dst_size* and *ratio*. Set `remove_on_error` to remove the destination directory
    in case of error.

    This function will return a dictionary containing *start_date*, *elapsed_time* and *src_size*.
    At the end of the copy, if the size of the destination directory is not equal to the source
    then a `IOError` is raised.
    """
    try:
        start_date, start_time = datetime_now(), time.time()
        src_size, dst_size = get_size(source_path), 0

        # Recursive copy of a directory of files
        for src_root, _, filenames in os.walk(source_path):
            for filename in filenames:
                dst_root = src_root.replace(source_path, destination_path)
                src_path = os.path.join(src_root, filename)
                dst_path = os.path.join(dst_root, filename)

                # Initialize block-based copy
                makedirs(os.path.dirname(dst_path))
                block_size = 1024 * 1024
                src_file = open(src_path, 'rb')  # pylint: disable=consider-using-with
                dst_file = open(dst_path, 'wb')  # pylint: disable=consider-using-with

                # Block-based copy loop
                block_pos = prev_ratio = prev_time = 0
                while True:
                    block = src_file.read(block_size)
                    try:
                        ratio = float(dst_size) / src_size
                        ratio = 0.0 if ratio < 0.0 else 1.0 if ratio > 1.0 else ratio
                    except ZeroDivisionError:
                        ratio = 1.0
                    elapsed_time = time.time() - start_time

                    # Update status of job only if delta time or delta ratio is sufficient
                    if (
                        progress_callback is not None
                        and ratio - prev_ratio > ratio_delta
                        and elapsed_time - prev_time > time_delta
                    ):
                        prev_ratio = ratio
                        prev_time = elapsed_time
                        eta_time = int(elapsed_time * (1.0 - ratio) / ratio) if ratio > 0 else 0
                        progress_callback(
                            start_date,
                            elapsed_time,
                            eta_time,
                            src_size,
                            dst_size,
                            ratio)

                    block_length = len(block)
                    block_pos += block_length
                    dst_size += block_length

                    if not block:
                        break  # End of input reached

                    dst_file.write(block)

                # FIXME maybe a finally block for that
                src_file.close()
                dst_file.close()

        # Output directory sanity check
        if check_size:
            dst_size = get_size(destination_path)
            if dst_size != src_size:
                raise IOError(f'Destination size does not match source ({src_size} vs {dst_size})')

        elapsed_time = time.time() - start_time
        return {'start_date': start_date, 'elapsed_time': elapsed_time, 'src_size': src_size}
    except Exception:
        if remove_on_error:
            shutil.rmtree(destination_path, ignore_errors=True)
        raise


def remove(path, recursive=False):
    """
    Remove a file/directory (which may not exists) without throwing an exception.
    Returns True if operation is successful, False if file/directory not found and re-raise any
    other type of exception.

    **Example usage**

    >>> with open('remove.example', 'w', encoding='utf-8') as f:
    ...     f.write('salut les pépés')
    15
    >>> remove('remove.example')
    True
    >>> remove('remove.example')
    False

    >>> for file_name in ('remove/a', 'remove/b/c', 'remove/d/e/f'):
    ...     _ = makedirs(os.path.dirname(file_name))
    ...     with open(file_name, 'w', encoding='utf-8') as f:
    ...         f.write('salut les pépés')
    15
    15
    15
    >>> remove('remove/d/e', recursive=True)
    True
    >>> remove('remove/d/e', recursive=True)
    False
    >>> from pytoolbox.unittest import asserts
    >>> asserts.raises(OSError, remove, 'remove/b')
    >>> remove('remove', recursive=True)
    True
    """
    try:
        try:
            os.remove(path)
            return True
        except Exception as e:
            # Is a directory and recursion is allowed
            if recursive and (
                isinstance(e, OSError) and e.errno == errno.EISDIR  # pylint:disable=no-member
                or PermissionError is not None and isinstance(e, PermissionError)
            ):
                shutil.rmtree(path)
                return True
            raise  # Re-raise exception if a different error occurred
    except OSError as e:
        # File does not exist
        if e.errno == errno.ENOENT:
            return False
        raise  # Re-raise exception if a different error occurred


def symlink(source, link_name):
    """
    Symlink a file/directory (which may already exists) without throwing an exception. Returns True
    if operation is successful, False if found & target is `link_name` and re-raise any other type
    of exception.

    **Example usage**

    >>> a = remove('/tmp/does_not_exist')
    >>> a = remove('/tmp/does_not_exist_2')
    >>> a = remove('/tmp/link_etc')
    >>> a = remove(os.path.expanduser('~/broken_link'))

    Creating a symlink named /etc does fail - /etc already exist but does not refer to /home:

    >>> from pytoolbox.unittest import asserts
    >>> asserts.raises(OSError, symlink, '/home', '/etc')

    Symlinking /etc to itself only returns that nothing changed:

    >>> symlink('/etc', '/etc')
    False

    Creating a symlink to an existing file has the following behaviour:

    >>> symlink('/etc', '/tmp/link_etc')
    True
    >>> symlink('/etc', '/tmp/link_etc')
    False
    >>> asserts.raises(OSError, symlink, '/etc/does_not_exist', '/tmp/link_etc')
    >>> asserts.raises(OSError, symlink, '/home', '/tmp/link_etc')

    Creating a symlink to a non existing has the following behaviour:

    >>> symlink('~/does_not_exist', '~/broken_link')
    True
    >>> symlink('~/does_not_exist', '~/broken_link')
    False
    >>> asserts.raises(OSError, symlink, '~/does_not_exist_2', '~/broken_link')
    >>> asserts.raises(OSError, symlink, '/home', '~/broken_link')
    >>> os.remove('/tmp/link_etc')
    >>> os.remove(os.path.expanduser('~/broken_link'))
    """
    try:
        source = os.path.expanduser(source)
        link_name = os.path.expanduser(link_name)
        os.symlink(source, link_name)
        return True
    except OSError as e1:  # pylint:disable=invalid-name
        # File exists
        if e1.errno == errno.EEXIST:
            try:
                if os.path.samefile(source, link_name):
                    return False
            except OSError as e2:  # pylint:disable=invalid-name
                # Handle broken symlink that point to same target
                target = os.path.expanduser(os.readlink(link_name))
                if e2.errno == errno.ENOENT:
                    if target == source:
                        return False
                    raise OSError(errno.EEXIST, 'File exists')  # pylint:disable=raise-missing-from
                raise
        raise  # Re-raise exception if a different error occurred


def to_user_id(user):
    """
    Return user ID.

    **Example usage**

    >>> to_user_id('root')
    0
    >>> to_user_id(0)
    0
    >>> to_user_id(None)
    -1
    """
    if isinstance(user, str):
        return pwd.getpwnam(user).pw_uid
    return -1 if user is None else user


def to_group_id(group):
    """
    Return group ID.

    **Example usage**

    >>> to_group_id('root')
    0
    >>> to_group_id(0)
    0
    >>> to_group_id(None)
    -1
    """
    if isinstance(group, str):
        return grp.getgrnam(group).gr_gid
    return -1 if group is None else group


class TempStorage(object):
    """
    Temporary storage handling made easy.

    **Example usage**

    As a context manager:

    >>> import os
    >>> with TempStorage() as tmp:
    ...     directory = tmp.create_tmp_directory()
    ...     os.path.isdir(directory)
    True
    >>> os.path.isdir(directory)
    False
    """
    def __init__(self, root=None):
        self.root = root or tempfile.gettempdir()
        self._path_to_key = {}
        self._paths_by_key = collections.defaultdict(set)

    def __enter__(self):
        return self

    def __exit__(self, kind, value, traceback):
        self.remove_all()

    def create_tmp_directory(self, path='tmp-{uuid}', key=None, user=None, group=None):
        """
        **Example usage**

        >>> import os
        >>> tmp = TempStorage()
        >>> directory = tmp.create_tmp_directory()
        >>> os.path.isdir(directory)
        True
        >>> tmp.remove_all()
        """
        directory = os.path.join(self.root, path.format(uuid=uuid.uuid4().hex))
        self._path_to_key[directory] = key
        self._paths_by_key[key].add(directory)
        makedirs(directory)
        chown(directory, user, group, recursive=True)
        return directory

    def create_tmp_file(
        self,
        path='tmp-{uuid}',
        extension=None,
        encoding='utf-8',
        key=None,
        user=None,
        group=None,
        return_file=True
    ):
        """
        **Example usage**

        >>> import os
        >>> tmp = TempStorage()
        >>> path = tmp.create_tmp_file(encoding=None, return_file=False)
        >>> os.path.isfile(path)
        True
        >>> with tmp.create_tmp_file(extension='txt') as f:
        ...     assert os.path.isfile(f.name)
        ...     length = f.write('Je suis une théière')
        ...     path = f.name
        >>> length
        19
        >>> open(path).read()
        'Je suis une théière'
        >>> tmp.remove_all()
        """
        mode = 'w' if encoding else 'wb'
        path = os.path.join(
            self.root,
            path.format(uuid=uuid.uuid4().hex) + (f'.{extension}' if extension else ''))

        self._path_to_key[path] = key
        self._paths_by_key[key].add(path)

        with open(path, mode, encoding=encoding):
            pass

        chown(path, user, group)

        if return_file:
            return open(path, mode, encoding=encoding)  # pylint: disable=consider-using-with
        return path

    def remove_by_path(self, path):
        """
        **Example usage**

        >>> from pytoolbox.unittest import asserts
        >>> tmp = TempStorage()
        >>> directory = tmp.create_tmp_directory()
        >>> tmp.remove_by_path(directory)
        >>> with asserts.raises(KeyError):
        ...     tmp.remove_by_path(directory)
        >>> with asserts.raises(KeyError):
        ...     tmp.remove_by_path('random-path')
        """
        key = self._path_to_key[path]
        remove(path, recursive=True)
        del self._path_to_key[path]
        self._paths_by_key[key].remove(path)

    def remove_by_key(self, key=None):
        """
        **Example usage**

        >>> import os
        >>> tmp = TempStorage()
        >>> d1 = tmp.create_tmp_directory()
        >>> d2 = tmp.create_tmp_directory(key=10)
        >>> tmp.remove_by_key(10)
        >>> os.path.isdir(d1)
        True
        >>> os.path.isdir(d2)
        False
        >>> tmp.remove_by_key()
        >>> os.path.isdir(d1)
        False
        """
        paths = self._paths_by_key[key]
        for path in copy.copy(paths):
            remove(path, recursive=True)
            paths.remove(path)
            del self._path_to_key[path]
        del self._paths_by_key[key]

    def remove_all(self):
        """
        **Example usage**

        >>> import os
        >>> tmp = TempStorage()
        >>> d1 = tmp.create_tmp_directory()
        >>> d2 = tmp.create_tmp_directory(key=10)
        >>> tmp.remove_all()
        >>> os.path.isdir(d1)
        False
        >>> os.path.isdir(d2)
        False
        >>> tmp.remove_all()
        """
        for key in self._paths_by_key.copy().keys():
            self.remove_by_key(key)


__all__ = _all.diff(globals())



import setuptools, sys


class Disabled(setuptools.Command):  # pylint:disable=duplicate-code

    description = 'Do not run this.'
    user_options = [('dummy=', 'd', 'dummy option to make setuptools happy')]

    def initialize_options(self):
        """Initialize options."""

    def finalize_options(self):
        """Finalize options."""

    def run(self):
        sys.exit('This command is disabled!')
